import keras
from keras import random, ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.MixUp")
class MixUp(keras.layers.Layer):
    """
    MixUp is a regularization layer proposed by Hongyi Zhang et al.
    in the paper,
    mixup: BEYOND EMPIRICAL RISK MINIMIZATION

    It was originally proposed for image data but here it has been
    adapted for Tabular data.

    Reference(s):
        https://arxiv.org/abs/1710.09412

    Args:
        alpha: float, Parameter for the Beta distribution to sample
            `lambda_` from which is used to interpolate samples.
        lambda_seed: int, seed for sampling `lambda_` value from beta
            distribution. Defaults to 1337
        shuffle_seed: int, seed for randomly shuffling inputs.
            Defaults to 1999
    """
    def __init__(self,
                 alpha: float = 1.,
                 lambda_seed: int = 1337,
                 shuffle_seed: int = 1999,
                 **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.lambda_seed = lambda_seed
        self._lambda_seed_gen = random.SeedGenerator(self.lambda_seed)
        self.shuffle_seed = shuffle_seed
        self._shuffle_seed_gen = random.SeedGenerator(self.shuffle_seed)

    def build(self, input_shape):
        # there's nothing to build lol
        pass

    def call(self, inputs):
        # Sample lambda_
        lambda_ = ops.squeeze(random.beta(shape=(1,),
                                          alpha=self.alpha,
                                          beta=self.alpha,
                                          seed=self._lambda_seed_gen))
        # For each data sample select a partner to mix it with at random.
        # To efficiently achieve this, we can just shuffle the data
        random_partners = random.shuffle(inputs,
                                         axis=0,
                                         seed=self._shuffle_seed_gen)

        inputs_mixedup = (lambda_ * inputs) + (1 - lambda_) * random_partners
        return inputs_mixedup

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            "alpha": self.alpha,
            "lambda_seed": self.lambda_seed,
            "shuffle_seed": self.shuffle_seed,
        })
        return config



from keras.src.testing.test_case import TestCase
from keras import ops, random
from teras._src.layers.transformer.feedforward import TransformerFeedForward


class TransformerFeedForwardTest(TestCase):
    def test_valid_output_shape(self):
        input_shape = (4, 5, 16)
        inputs = random.normal(shape=input_shape)
        encoder_layer = TransformerFeedForward(embedding_dim=16,
                                               hidden_dim=64
                                               )
        outputs = encoder_layer(inputs)
        self.assertEqual(ops.shape(outputs), input_shape)



import os

import keras
from keras import ops
from keras.src.testing.test_case import TestCase

from teras._src.models.gans.gain.discriminator import GAINDiscriminator


class GAINDiscriminatorTest(TestCase):
    def setUp(self):
        self.latent_inputs = ops.ones((8, 16))

    def test_valid_call(self):
        discriminator = GAINDiscriminator(data_dim=7)
        outputs = discriminator(self.latent_inputs)

    def test_valid_output_shape(self):
        discriminator = GAINDiscriminator(data_dim=7)
        outputs = discriminator(self.latent_inputs)
        assert ops.shape(outputs) == (8, 7)

    def test_save_and_load(self):
        discriminator = GAINDiscriminator(data_dim=7)
        outputs_original = discriminator(self.latent_inputs)
        save_path = os.path.join(self.get_temp_dir(),
                                 "gain_discriminator.keras")
        discriminator.save(save_path)
        reloaded_model = keras.models.load_model(save_path)
        outputs_reloaded = reloaded_model(self.latent_inputs)
        self.assertAllClose(ops.convert_to_numpy(outputs_original),
                            ops.convert_to_numpy(outputs_reloaded))



from keras import ops, random
from teras._src.layers.categorical_extraction import CategoricalExtraction
from keras.src.testing.test_case import TestCase
from teras._src.utils import compute_cardinalities


class CategoricalExtractionTest(TestCase):
    def setUp(self):
        self.input_batch = ops.array(
            [random.normal((16,)),
             ops.repeat(ops.array([1., 2., 3., 4.]), 4),
             random.normal((16,)),
             ops.repeat(ops.array([10., 20.]), 8),
             random.normal((16,))])
        self.input_batch = ops.transpose(self.input_batch)
        self.categorical_idx = [1, 3]
        self.categorical_cardinalities = [4, 2]

    def test_valid_output_shape(self):
        extraction = CategoricalExtraction(
            categorical_idx=self.categorical_idx)
        categorical_features = extraction(self.input_batch)
        self.assertEqual((16, 2), ops.shape(categorical_features))

    def test_valid_output(self):
        extraction = CategoricalExtraction(
            categorical_idx=self.categorical_idx)
        categorical_features = extraction(self.input_batch)
        categorical_features = ops.convert_to_numpy(categorical_features)
        cards = compute_cardinalities(
            categorical_features,
            categorical_idx=list(range(categorical_features.shape[1])))
        cards = list(cards)
        self.assertEqual(self.categorical_cardinalities, cards)



from keras import random, ops
from teras._src.layers.tabnet.attentive_transformer import TabNetAttentiveTransformer
from keras.src.testing.test_case import TestCase


class TabNetAttentiveTransformerTest(TestCase):
    def setUp(self):
        self.batch_size = 16
        self.data_dim = 5
        self.hidden_representations = random.normal((self.batch_size, 8))

    def test_valid_output_shape(self):
        layer = TabNetAttentiveTransformer(data_dim=self.data_dim,
                                           batch_momentum=0.99)
        prior_scales = random.normal((self.batch_size, self.data_dim))
        outputs = layer(self.hidden_representations,
                        prior_scales=prior_scales)
        self.assertEqual(ops.shape(outputs),
                         (self.batch_size, self.data_dim))

    def test_valid_output_shape_with_prior_scales(self):
        layer = TabNetAttentiveTransformer(data_dim=self.data_dim,
                                           batch_momentum=0.99)
        prior_scales = random.normal((self.batch_size, self.data_dim))
        outputs = layer(self.hidden_representations,
                        prior_scales=prior_scales)
        self.assertEqual(ops.shape(outputs),
                         (self.batch_size, self.data_dim))



import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.SAINTPretrainer")
class SAINTPretrainer(backend.models.SAINTPretrainer):
    """
    SAINTPretrainer as proposed in the paper,
    "SAINT: Improved Neural Networks for Tabular Data".

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        model: keras.Model, instance of `SAINTBackbone` model to pretrain
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use the value `0` as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, dimensionality of the embeddings being used
             in the model.
        cutmix_probability: float, used by the `CutMix` layer in
            generation of mask that is used to mix samples together.
            Defaults to 0.3
        mixup_alpha: float, used by the `MixUp` layer in sampling from the
            `Beta` distribution which is then used to interpolate samples.
            Defaults to 1.
        temperature: float, used in the computation of the
            `contrastive_loss` to scale logits. Defaults to 0.7
        lambda_: float, acts as a weight when adding the contrastive loss
            and the denoising loss together.
            `loss = constrastive_loss + lambda_ * denoising_loss`
            Defaults to 10.
        lambda_c: float, used in the computation of the contrastive
            loss. Similar to `lambda_` is helps combined two sub-losses
            within the contrastive loss. Defaults to 0.5
        seed: int, seed used in random sampling and shuffling etc.
            It helps make the model behavior more deterministic.
            Defaults to (you guessed it) 1337.
    """
    def __init__(self,
                 model: keras.Model,
                 cardinalities: list,
                 embedding_dim: int,
                 cutmix_probability: float = 0.3,
                 mixup_alpha: float = 1.,
                 temperature: float = 0.7,
                 lambda_: float = 10.,
                 lambda_c: float = 0.5,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         cardinalities=cardinalities,
                         embedding_dim=embedding_dim,
                         cutmix_probability=cutmix_probability,
                         mixup_alpha=mixup_alpha,
                         temperature=temperature,
                         lambda_=lambda_,
                         lambda_c=lambda_c,
                         seed=seed,
                         **kwargs)



import keras
import numpy as np
from keras import ops
from teras._src.api_export import teras_export


_VALID_JOIN_METHODS = ['concat', 'add']


@teras_export("teras.layers.TabTransformerColumnEmbedding")
class TabTransformerColumnEmbedding(keras.layers.Layer):
    """
    Column Embedding layer as proposed in the
    "TabTransformer: Tabular Data Modeling Using Contextual Embeddings".

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use any value <=0 as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, dimensionality of the embeddings
        use_shared_embedding: bool, whether to use the shared embeddings.
            Defaults to `True`.
            If `False`, this layer will be effectively equivalent to a
            `keras.layers.Embedding` layer.
        shared_embedding_dim: int, dimensionality of the shared embeddings.
            Shared embeddings are the embeddings of the unique column
            identifiers, which according to the paper, help the model
            distinguish categories of one feature from the other.
            By default, its value is set to `embedding_dim / 8` as this
            setup is the most superior in the results presented by the
            authors.
        join_method: str, one of ['concat', 'add'] method to join the
            shared embeddings with feature embeddings.
            Defaults to 'concat', which is the recommended method,
            in which shared embeddings of `shared_embedding_dim` are
            concatenated with `embedding_dim - shared_embedding_dim`
            dimension feature embeddings.
            In 'add', shared embeddings have the same dimensions as the
            features, i.e. the `embedding_dim` and they are element-wise
            added to the features.
    """
    def __init__(self,
                 cardinalities: list,
                 embedding_dim: int,
                 use_shared_embedding: bool = True,
                 shared_embedding_dim: int = None,
                 join_method: str = "concat",
                 **kwargs):
        super().__init__(**kwargs)
        if join_method not in _VALID_JOIN_METHODS:
            raise ValueError(
                f"`join_method` must be one of {_VALID_JOIN_METHODS}, "
                f"but received, {join_method}")
        self.cardinalities = cardinalities
        self.embedding_dim = embedding_dim
        self.use_shared_embedding = use_shared_embedding
        self.shared_embedding_dim = (
            shared_embedding_dim if shared_embedding_dim else
            embedding_dim // 8)
        self.join_method = join_method
        # indices of categorical features
        self._cardinalities_arr = np.array(self.cardinalities)
        self._categorical_idx = np.flatnonzero(
            self._cardinalities_arr != 0)

        categorical_cardinalities = self._cardinalities_arr[self._categorical_idx]

        # The paper assumes each categorical feature has
        # num_categories + 1 embeddings where additional embedding
        # corresponds to a missing value - but we'll just stick to using
        # one extra embedding for the whole embedding table
        num_special_tokens = 1

        # We create a one big lookup table of embeddings for all
        # the categories/classes of categorical features. And we assume
        # the categorical features to be ordinally encoded, hence
        # multiple features will have common values like, [0, 1, 2, ...]
        # So, to distinguish the common values of one feature from the
        # other we create an offset array to offset the values.
        # Additionally, we insert 1 to the leftmost position to reserve
        # the value of 0 for special purpose of 'missing value'.
        categorical_cardinalities = ops.pad(
            categorical_cardinalities,
            (1, 0),
            constant_values=num_special_tokens)
        self._categories_offset = ops.cumsum(
            categorical_cardinalities)[:-1]
        self._categories_offset = ops.cast(self._categories_offset,
                                           "float32")
        self._total_tokens = sum(self.cardinalities) + num_special_tokens

    def build(self, input_shape=None):
        num_categorical_features = len(self._categorical_idx)
        feature_embedding_dim = self.embedding_dim
        shared_embedding_dim = self.shared_embedding_dim

        if self.use_shared_embedding:
            if self.join_method == "concat":
                feature_embedding_dim -= shared_embedding_dim
            # if user specifies any shared embedding dimensions, but the
            # join method is set to add, then shared embeddings will be
            # the same dimensions as the feature embedding and user's
            # supplied value will be overwritten
            else:
                shared_embedding_dim = feature_embedding_dim

        self.feature_embedding = self.add_weight(
            shape=(self._total_tokens, feature_embedding_dim),
            initializer="random_normal",
            trainable=True,
            name="feature_embedding",
        )

        if self.use_shared_embedding:
            self.shared_embedding = self.add_weight(
                shape=(num_categorical_features,
                       shared_embedding_dim),
                initializer="random_normal",
                trainable=True,
                name="shared_embedding"
            )
        self.built = True

    def call(self, inputs):
        _dtype = inputs.dtype
        _batch_size = ops.shape(inputs)[0]
        inputs = ops.take(inputs, indices=self._categorical_idx,
                          axis=1)
        inputs += self._categories_offset
        inputs = ops.cast(inputs, "int32")
        embeddings = ops.take(self.feature_embedding,
                              indices=inputs,
                              axis=0)
        embeddings = ops.cast(embeddings, dtype=_dtype)
        if self.use_shared_embedding:
            if self.join_method == "concat":
                _shared_emb = ops.tile(self.shared_embedding,
                                       (_batch_size, 1, 1))
                embeddings = ops.concatenate([_shared_emb, embeddings],
                                             axis=-1)
            else:
                embeddings += self.shared_embedding
        return embeddings

    def compute_output_shape(self, input_shape):
        return input_shape + (self.embedding_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "cardinalities": self.cardinalities,
            "embedding_dim": self.embedding_dim,
            "use_shared_embedding": self.use_shared_embedding,
            "shared_embedding_dim": self.shared_embedding_dim,
            "join_method": self.join_method
        })
        return config



def dataset_type(x):
    if str(type(x)).__contains__("tensorflow"):
        return "tensorflow_dataset"
    elif str(type(x)).__contains__("DataLoader"):
        return "torch_dataloader"
    # elif str(type(x)).__contains__("numpy.ndarray"):
    #     return "numpy_ndarray"
    else:
        return "not_supported"



import keras
from teras._src.api_export import teras_export


@teras_export("teras.layers.LayerList")
class LayerList(keras.layers.Layer):
    """
    LayerList is a list of layers, but is also a layer itself.
    If you know what that means, great. if not, well it's alright.
    you'll get there, one day, sooner or later. just keep pushing. keep
    learning. never give up. you got this, king/queen!

    Args:
        layers: list, list of Keras layers
        sequential: bool, whether to build layers sequentially.
            Set it to True ONLY when each layer in the list is
            applied one after the other in a sequential manner.
            Otherwise, set it to False.
            If sequential, each layer is built using the output
            shape of the previous layer, with first layer being
            built with the `input_shape` argument to the `build` method.
            Otherwise, each layer is built using the `input_shape`
            argument received.
    """
    def __init__(self,
                 layers: list,
                 sequential: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        self.layers = layers
        self.sequential = sequential

    def __getitem__(self, item):
        return self.layers[item]

    def __len__(self):
        return len(self.layers)

    def compute_output_shape(self, input_shape):
        if not self.sequential:
            return self.layers[-1].compute_output_shape(input_shape)
        else:
            output_shape = input_shape
            for layer in self.layers:
                output_shape = layer.compute_output_shape(output_shape)
            return output_shape

    def build(self, input_shape):
        """
        Build.

        Args:
            input_shape: i wonder what that means
        """
        for layer in self.layers:
            layer.build(input_shape)
            if self.sequential:
                input_shape = layer.compute_output_shape(input_shape)

    def call(self, inputs):
        if self.sequential:
            x = inputs
            for layer in self.layers:
                x = layer(x)
            return x
        else:
            raise NotImplemented(
                "`LayerList` doesn't provide a `call` method for "
                "non-sequential list of layers."
            )

    def get_config(self):
        config = super().get_config()
        serialized_layers = [keras.layers.serialize(layer)
                             for layer in self.layers]
        config.update({
            "layers": serialized_layers,
            "sequential": self.sequential
        })
        return config

    def from_config(cls, config):
        deserialized_layers = [keras.layers.deserialize(layer)
                               for layer in config.pop("layers")]
        return cls(deserialized_layers,
                   **config)



import torch
import keras
from keras import random, ops
from keras.backend import floatx
from teras._src.backend.common.models.gans.pcgain import BasePCGAIN


class PCGAIN(BasePCGAIN):
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 classifier: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 200.,
                 beta: float = 100.,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         classifier=classifier,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         beta=beta,
                         **kwargs)

    def train_step(self, data):
        # data is a tuple of x_generator and x_discriminator batches
        # drawn from the dataset. The reason behind generating two separate
        # batches of data at each step is that it's how GAIN's algorithm works
        x_gen, x_disc = data

        self.zero_grad()

        # =========================
        # Train the discriminator
        # =========================
        # Create mask
        mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
        # replace nans with 0.
        x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
        # Sample noise
        z = random.uniform(shape=ops.shape(x_disc), minval=0., maxval=0.01,
                           seed=self._seed_gen)
        # Sample hint vectors
        hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors = hint_vectors * mask
        # Combine random vectors with original data
        x_disc = x_disc * mask + (1 - mask) * z
        x_generated = self.generator(ops.concatenate([x_disc, mask], axis=1))
        # Combine generated samples with original data
        x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)
        mask_pred = self.discriminator(
            ops.concatenate([x_hat_disc, hint_vectors], axis=1))
        d_loss = self.compute_discriminator_loss(mask, mask_pred)

        d_loss.backward()
        gradients = [v.value.grad for v in self.discriminator.trainable_weights]
        with torch.no_grad():
            self.discriminator_optimizer.apply(
                gradients,
                self.discriminator.trainable_weights)

        # =====================
        # Train the generator
        # =====================
        mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
        x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
        z = random.uniform(shape=ops.shape(x_gen), minval=0., maxval=0.01,
                           seed=self._seed_gen)
        hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors = hint_vectors * mask
        x_gen = x_gen * mask + (1 - mask) * z

        x_generated = self.generator(
            ops.concatenate([x_gen, mask], axis=1))
        # Combine generated samples with original/observed data
        x_hat = (x_generated * (1 - mask)) + (x_gen * mask)
        classifier_pred = self.classifier(x_hat)
        mask_pred = self.discriminator(
            ops.concatenate([x_hat, hint_vectors], axis=1))
        g_loss = self.compute_generator_loss(
            x=x_gen,
            x_generated=x_generated,
            mask=mask,
            mask_pred=mask_pred,
            classifier_pred=classifier_pred,
            alpha=self.alpha,
            beta=self.beta,
        )
        g_loss.backward()
        gradients = [v.value.grad for v in self.generator.trainable_weights]
        with torch.no_grad():
            self.generator_optimizer.apply(
                gradients,
                self.generator.trainable_weights)

        # Update custom tracking metrics
        self.generator_loss_tracker.update_state(g_loss)
        self.discriminator_loss_tracker.update_state(d_loss)

        logs = {m.name: m.result() for m in self.metrics}
        return logs



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.CLSToken")
class CLSToken(keras.layers.Layer):
    """
    CLS Token layer that makes it possible to append CLS token embedding
    to the input embeddings in the sequential or functional models.

    The idea of CLS token was introduced in the "BERT: Pre-training of
    Deep Bidirectional Transformers for Language Understanding" paper.

    Reference(s):
        https://arxiv.org/abs/1810.04805

    Args:
        embedding_dim: int, dimensionality of the input embeddings

    Shapes:
        Input Shape: `(batch_size, num_features, embedding_dim)`
        Output Shape: `(batch_size, num_features + 1, embedding_dim)`
    """
    def __init__(self,
                 embedding_dim: int,
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim

    def build(self, input_shape=None):
        self.cls_token = self.add_weight(
            shape=(1, self.embedding_dim),
            initializer="random_normal",
        )

    def call(self, inputs):
        # TODO Remove the call to `convert_to_tensor` as soon as Keras
        #  fixes `broadcast_to` method for JAX backend
        token_broadcasted = ops.broadcast_to(
            ops.convert_to_tensor(self.cls_token),
            shape=(ops.shape(inputs)[0], *ops.shape(self.cls_token)))
        return ops.concatenate([token_broadcasted, inputs],
                               axis=1)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[1] + 1, input_shape[2])

    def get_config(self):
        config = super().get_config()
        config.update({
            "embedding_dim": self.embedding_dim
        })
        return config



import tensorflow as tf
import keras
from keras import random, ops
from keras.backend import floatx
from teras._src.backend.common.models.gans.gain import BaseGAIN


class GAIN(BaseGAIN):
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 100.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         seed=seed,
                         **kwargs)

    def train_step(self, data):
        # data is a tuple of x_generator and x_discriminator batches
        # drawn from the dataset. The reason behind generating two separate
        # batches of data at each step is that it's how GAIN's algorithm works
        x_gen, x_disc = data

        # =========================
        # Train the discriminator
        # =========================
        # Create mask
        mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
        # replace nans with 0.
        x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
        # Sample noise
        z = random.uniform(shape=ops.shape(x_disc), minval=0.,
                           maxval=0.01, seed=self._seed_gen)
        # Sample hint vectors
        hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors *= mask
        # Combine random vectors with original data
        x_disc = x_disc * mask + (1 - mask) * z
        x_generated = self.generator(ops.concatenate([x_disc, mask], axis=1))
        # Combine generated samples with original data
        x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)
        with tf.GradientTape() as tape:
            mask_pred = self.discriminator(
                ops.concatenate([x_hat_disc, hint_vectors], axis=1))
            d_loss = self.compute_discriminator_loss(mask, mask_pred)
        gradients = tape.gradient(d_loss, self.discriminator.trainable_weights)
        self.discriminator_optimizer.apply(gradients,
                                           self.discriminator.trainable_weights)

        # =====================
        # Train the generator
        # =====================
        mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
        x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
        z = random.uniform(shape=ops.shape(x_gen), minval=0.,
                           maxval=0.01, seed=self._seed_gen)
        hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors *= mask
        x_gen = x_gen * mask + (1 - mask) * z

        with tf.GradientTape() as tape:
            x_generated = self.generator(
                ops.concatenate([x_gen, mask], axis=1))
            # Combine generated samples with original/observed data
            x_hat = (x_generated * (1 - mask)) + (x_gen * mask)
            mask_pred = self.discriminator(
                ops.concatenate([x_hat, hint_vectors], axis=1))
            g_loss = self.compute_generator_loss(
                x=x_gen,
                x_generated=x_generated,
                mask=mask,
                mask_pred=mask_pred,
                alpha=self.alpha
            )
        gradients = tape.gradient(g_loss, self.generator.trainable_weights)
        self.generator_optimizer.apply(gradients,
                                       self.generator.trainable_weights)

        # Update custom tracking metrics
        self.generator_loss_tracker.update_state(g_loss)
        self.discriminator_loss_tracker.update_state(d_loss)

        logs = {m.name: m.result() for m in self.metrics}
        return logs



from keras import ops
from teras._src.activations import (glu,
                               geglu,
                               gumbel_softmax,
                               sparsemax)


# ================== glu tests ====================
def test_glu_valid_call():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = glu(inputs)


def test_glu_output_shape():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = glu(inputs)
    assert len(ops.shape(outputs)) == 2
    assert ops.shape(outputs)[0] == 8
    assert ops.shape(outputs)[1] == 2    # glu halves the dimensions


# ================== geglu tests ====================
def test_geglu_valid_call():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = geglu(inputs)


def test_geglu_output_shape():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = geglu(inputs)
    assert len(ops.shape(outputs)) == 2
    assert ops.shape(outputs)[0] == 8
    assert ops.shape(outputs)[1] == 2    # geglu halves the dimensions


# ================== gumbel_softmax tests ====================
def test_gumbel_softmax_valid_call():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = gumbel_softmax(inputs)


def test_gumbel_softmax_output_shape():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = gumbel_softmax(inputs)
    assert len(ops.shape(outputs)) == 2
    assert ops.shape(outputs)[0] == 8
    assert ops.shape(outputs)[1] == 4


# ================== sparsemax tests ====================
def test_sparsemax_valid_call():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = gumbel_softmax(inputs)


def test_sparsemax_output_shape():
    inputs = ops.ones((8, 4), dtype="float32")
    outputs = gumbel_softmax(inputs)
    assert len(ops.shape(outputs)) == 2
    assert ops.shape(outputs)[0] == 8
    assert ops.shape(outputs)[1] == 4



import numpy as np


_DEFAULT_NUMPY_FLOATX = np.float32


def get_default_numpy_floatx():
    return _DEFAULT_NUMPY_FLOATX


def set_default_numpy_floatx(dtype):
    global _DEFAULT_NUMPY_FLOATX
    _DEFAULT_NUMPY_FLOATX = dtype



import jax
import keras
from keras import random, ops
from keras.backend import floatx
from teras._src.backend.jax.models.gans.jaxgan import JAXGAN
from teras._src.backend.common.models.gans.gain import BaseGAIN


class GAIN(JAXGAN, BaseGAIN):
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 100.,
                 seed: int = 1337,
                 **kwargs
                 ):
        JAXGAN.__init__(self,
                        generator=generator,
                        discriminator=discriminator)
        BaseGAIN.__init__(self,
                          generator=generator,
                          discriminator=discriminator,
                          hint_rate=hint_rate,
                          alpha=alpha,
                          seed=seed,
                          **kwargs)
        self._prng_key = jax.random.PRNGKey(seed)

    def generator_compute_loss_and_updates(
            self,
            generator_trainable_vars,
            generator_non_trainable_vars,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_gen,
            hint_vectors,
            mask,
            training=False
    ):
        x_generated, generator_non_trainable_vars = self.generator.stateless_call(
            generator_trainable_vars,
            generator_non_trainable_vars,
            ops.concatenate([x_gen, mask], axis=1),
            training=training,
        )
        x_hat = (x_generated * (1 - mask)) + (x_gen * mask)
        mask_pred, _ = (
            self.discriminator.stateless_call(
                discriminator_trainable_vars,
                discriminator_non_trainable_vars,
                ops.concatenate([x_hat, hint_vectors], axis=1),
                training=training,
            )
        )
        loss = self.compute_generator_loss(
            x=x_gen,
            x_generated=x_generated,
            mask=mask,
            mask_pred=mask_pred,
            alpha=self.alpha
        )
        return loss, (generator_non_trainable_vars,)

    def discriminator_compute_loss_and_updates(
            self,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_hat_disc,
            hint_vectors,
            mask,
            training=False
    ):
        mask_pred, discriminator_non_trainable_vars = (
            self.discriminator.stateless_call(
                discriminator_trainable_vars,
                discriminator_non_trainable_vars,
                ops.concatenate([x_hat_disc, hint_vectors], axis=1),
                training=training,
            )
        )
        loss = self.compute_discriminator_loss(mask, mask_pred)
        return loss, (discriminator_non_trainable_vars,)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        ) = state

        # GAIN contains SeedGenerator in its __init__ so it's the very first
        # non-trainable variable.
        seed = non_trainable_variables[0]

        (
            generator_state,
            discriminator_state,
        ) = self._parse_variables(trainable_variables,
                                  non_trainable_variables[1:],
                                  optimizer_variables)
        (
            generator_trainable_vars,
            generator_non_trainable_vars,
            generator_optimizer_vars
        ) = generator_state
        (
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            discriminator_optimizer_vars
        ) = discriminator_state

        # data is a tuple of x_generator and x_discriminator batches
        # drawn from the dataset. The reason behind generating two separate
        # batches of data at each step is that it's how GAIN's algorithm works
        x_gen, x_disc = data

        # =========================
        # Train the discriminator
        # =========================
        # Create mask
        mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
        # replace nans with 0.
        x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
        # Sample noise
        seed = jax.random.split(seed, 1)[0]
        z = random.uniform(shape=ops.shape(x_disc), minval=0.,
                           maxval=0.01, seed=seed)
        # Sample hint vectors
        seed = jax.random.split(seed, 1)[0]
        hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=seed)
        hint_vectors = hint_vectors * mask
        # Combine random vectors with original data
        x_disc = x_disc * mask + (1 - mask) * z

        x_generated, _ = self.generator.stateless_call(
            generator_trainable_vars,
            generator_non_trainable_vars,
            ops.concatenate([x_disc, mask], axis=1),
        )
        x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)
        disc_grad_fn = jax.value_and_grad(
            self.discriminator_compute_loss_and_updates,
            has_aux=True,
        )
        (d_loss, (discriminator_non_trainable_vars,)), grads = disc_grad_fn(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_hat_disc,
            hint_vectors,
            mask,
            training=True,
        )

        (
            discriminator_trainable_vars,
            discriminator_optimizer_vars
        ) = self.discriminator_optimizer.stateless_apply(
            discriminator_optimizer_vars,
            grads,
            discriminator_trainable_vars
        )

        # =====================
        # Train the generator
        # =====================
        mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
        x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
        seed = jax.random.split(seed, 1)[0]
        z = random.uniform(shape=ops.shape(x_gen), minval=0.,
                           maxval=0.01, seed=seed)
        seed = jax.random.split(seed, 1)[0]
        hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=seed)
        hint_vectors = hint_vectors * mask
        x_gen = x_gen * mask + (1 - mask) * z

        gen_grad_fn = jax.value_and_grad(
            self.generator_compute_loss_and_updates,
            has_aux=True,
        )
        (g_loss, (generator_non_trainable_vars,)), grads = gen_grad_fn(
            generator_trainable_vars,
            generator_non_trainable_vars,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_gen,
            hint_vectors,
            mask,
            training=True,
        )

        (
            generator_trainable_vars,
            generator_optimizer_vars
        ) = self.generator_optimizer.stateless_apply(
            generator_optimizer_vars,
            grads,
            generator_trainable_vars
        )

        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(new_metric_variables) + len(metric.variables)
                                    ]
            if metric.name == "generator_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    g_loss
                )
            elif metric.name == "discriminator_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    d_loss
                )
            else:
                continue
            logs[metric.name] = metric.stateless_result(this_metric_variables)
            new_metric_variables += this_metric_variables
        metrics_variables = new_metric_variables
        
        # Update seed value
        seed = seed
        state = (
            generator_trainable_vars + discriminator_trainable_vars,
            [seed] + generator_non_trainable_vars + 
            discriminator_non_trainable_vars,
            generator_optimizer_vars + discriminator_optimizer_vars,
            metrics_variables
        )
        return logs, state

    def test_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            metrics_variables,
        ) = state
        seed = non_trainable_variables[0]
        (
            generator_state,
            discriminator_state
        ) = self._parse_variables(trainable_variables,
                                  non_trainable_variables[1:])

        (
            generator_trainable_vars,
            generator_non_trainable_vars,
        ) = generator_state
        (
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
        ) = discriminator_state

        x_gen, x_disc = data

        # =========================
        # Test the discriminator
        # =========================
        # Create mask
        mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
        # replace nans with 0.
        x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
        # Sample noise
        seed = jax.random.split(seed, 1)[0]
        z = random.uniform(shape=ops.shape(x_disc), minval=0.,
                           maxval=0.01, seed=seed)
        # Sample hint vectors
        seed = jax.random.split(seed, 1)[0]
        hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=seed)
        hint_vectors *= mask
        # Combine random vectors with original data
        x_disc = x_disc * mask + (1 - mask) * z
        x_generated, _ = self.generator.stateless_call(
            generator_trainable_vars,
            generator_non_trainable_vars,
            ops.concatenate([x_disc, mask], axis=1))
        # Combine generated samples with original data
        x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)

        (
            d_loss,
            (discriminator_non_trainable_vars,)
        ) = self.discriminator_compute_loss_and_updates(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_hat_disc,
            hint_vectors,
            mask
        )

        # =====================
        # Test the generator
        # =====================
        mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
        x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
        seed = jax.random.split(seed, 1)[0]
        z = random.uniform(shape=ops.shape(x_gen), minval=0.,
                           maxval=0.01, seed=seed)
        seed = jax.random.split(seed, 1)[0]
        hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=seed)
        hint_vectors *= mask
        x_gen = x_gen * mask + (1 - mask) * z

        (
            g_loss,
            (generator_non_trainable_vars,)
        ) = self.generator_compute_loss_and_updates(
            generator_trainable_vars,
            generator_non_trainable_vars,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_gen,
            hint_vectors,
            mask,
        )

        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(new_metric_variables) + len(metric.variables)
                                    ]
            if metric.name == "generator_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    g_loss
                )
            elif metric.name == "discriminator_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    d_loss
                )
            else:
                continue
            logs[metric.name] = metric.stateless_result(this_metric_variables)
            new_metric_variables += this_metric_variables
        metrics_variables = new_metric_variables

        # Update seed
        seed = jax.random.split(seed, 1)[0]

        state = (
            generator_trainable_vars + discriminator_trainable_vars,
            [seed] + generator_non_trainable_vars +
            discriminator_non_trainable_vars,
            metrics_variables
        )
        return logs, state

    def predict_step(self, state, data):
        """
        Args:
            Transformed data.
        Returns:
            Imputed data that should be reverse transformed
            to its original form.
        """
        (
            trainable_variables,
            non_trainable_variables
        ) = state
        seed = non_trainable_variables[0]
        (
            generator_state,
            discriminator_state
        ) = self._parse_variables(trainable_variables,
                                  non_trainable_variables)
        (
            generator_trainable_vars,
            generator_non_trainable_vars
        ) = generator_state
        (
            discriminator_trainable_vars,
            discriminator_non_trainable_vars
        ) = discriminator_state
        if isinstance(data, tuple):
            data = data[0]
        data = ops.cast(data, floatx())
        # Create mask
        mask = 1. - ops.cast(ops.isnan(data), dtype=floatx())
        data = ops.where(ops.isnan(data), x1=0., x2=data)
        # Sample noise
        seed = jax.random.split(seed, 1)[0]
        z = random.uniform(ops.shape(data), minval=0.,
                           maxval=0.01, seed=seed)
        x = mask * data + (1 - mask) * z
        (
            imputed_data,
            generator_non_trainable_vars
        ) = self.generator.stateless_call(
            generator_trainable_vars,
            generator_non_trainable_vars,
            ops.concatenate([x, mask], axis=1))
        imputed_data = mask * data + (1 - mask) * imputed_data

        # Update seed
        seed = jax.random.split(seed, 1)[0]
        non_trainable_variables = (
                [seed] + generator_non_trainable_vars +
                discriminator_non_trainable_vars)
        return imputed_data, non_trainable_variables



from teras._src.layers.saint.reconstruction_head import SAINTReconstructionHead
from keras.src.testing.test_case import TestCase
from keras import ops, random
import numpy as np


class SAINTReconstructionHeadTest(TestCase):
    def setUp(self):
        self.input_batch = random.normal((8, 3, 16))
        # 0 represents the numerical features as required by the
        # `CategoricalEmbedding` layer
        self.cardinalities = [2, 10, 0]

    def test_valid_call(self):
        reconstruction_head = SAINTReconstructionHead(
            cardinalities=self.cardinalities,
            embedding_dim=16,
        )
        outputs = reconstruction_head(self.input_batch)

    def test_valid_output_shape(self):
        reconstruction_head = SAINTReconstructionHead(
            cardinalities=self.cardinalities,
            embedding_dim=16,
        )
        outputs = reconstruction_head(self.input_batch)
        self.assertEqual(ops.shape(outputs),
                         (8, sum(self.cardinalities) + 1))



import jax
import keras
from keras import random, ops
from teras._src.backend.jax.models.gans.jaxgan import JAXGAN
from teras._src.backend.common.models.gans.ctgan.ctgan import BaseCTGAN


class CTGAN(JAXGAN, BaseCTGAN):
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 metadata: dict,
                 latent_dim: int = 128,
                 seed: int = 1337,
                 **kwargs):
        JAXGAN.__init__(self,
                        generator=generator,
                        discriminator=discriminator,
                        **kwargs)
        BaseCTGAN.__init__(self,
                           generator=generator,
                           discriminator=discriminator,
                           metadata=metadata,
                           latent_dim=latent_dim,
                           seed=seed)

    @staticmethod
    def compute_generator_loss(x_generated,
                               y_pred_generated,
                               cond_vectors,
                               mask,
                               metadata):
        """
        Loss for the Generator model in the CTGAN architecture.

        Args:
            x_generated: Samples drawn from the input dataset
            y_pred_generated: Discriminator's output for the generated samples
            cond_vectors: Conditional vectors that are used for and with
                generated samples
            mask: Mask created during the conditional vectors generation step
            metadata: dict, metadata computed during the data transformation step.

        Returns:
            Generator's loss.
        """
        loss = []
        cross_entropy_loss = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=None)
        continuous_features_relative_indices = metadata["continuous"][
            "relative_indices_all"]
        features_relative_indices_all = metadata["relative_indices_all"]
        num_categories_all = metadata["categorical"]["num_categories_all"]
        # the first k features in the data are continuous which we'll ignore as
        # we're only concerned with the categorical features here
        offset = len(continuous_features_relative_indices)
        for i, index in enumerate(features_relative_indices_all[offset:]):
            start_idx = index
            slice_size = num_categories_all[i]
            logits = jax.lax.dynamic_slice_in_dim(
                x_generated, start_index=start_idx,
                slice_size=slice_size, axis=1
            )
            start_idx = i
            slice_size = num_categories_all[i]
            temp_cond_vector = jax.lax.dynamic_slice_in_dim(
                cond_vectors, start_index=start_idx,
                slice_size=slice_size, axis=1
            )
            labels = ops.argmax(temp_cond_vector, axis=1)
            ce_loss = cross_entropy_loss(y_pred=logits,
                                         y_true=labels
                                         )
            loss.append(ce_loss)
        loss = ops.stack(loss, axis=1)
        loss = ops.sum(
            loss * ops.cast(mask, dtype="float32")
                       ) / ops.cast(ops.shape(y_pred_generated)[0],
                                    dtype="float32")
        loss = -ops.mean(y_pred_generated) * loss
        return loss

    def generator_compute_loss_and_updates(
            self,
            generator_trainable_vars,
            generator_non_trainable_vars,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            noise,
            cond_vectors,
            mask,
            metadata
    ):
        (
            x_generated,
            generator_non_trainable_vars
        ) = self.generator.stateless_call(
            generator_trainable_vars,
            generator_non_trainable_vars,
            noise)
        x_generated = ops.concatenate(
            [x_generated, cond_vectors], axis=1)
        (
            y_pred_generated,
            discriminator_non_trainable_vars
        ) = self.discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_generated)
        loss = self.compute_generator_loss(x_generated, y_pred_generated,
                                           cond_vectors=cond_vectors, mask=mask,
                                           metadata=metadata)
        return loss, (generator_non_trainable_vars,)

    def discriminator_compute_grad_penalty_and_updates(
            self,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x,
            x_generated,
    ):
        (
            y_pred_generated,
            discriminator_non_trainable_vars
        ) = self.discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_generated)
        (
            y_pred_real,
            discriminator_non_trainable_vars
        ) = self.discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x)
        (
            grad_pen,
            discriminator_non_trainable_vars
        ) = self.discriminator.gradient_penalty(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x, x_generated)
        return grad_pen, (discriminator_non_trainable_vars,)

    def discriminator_compute_loss_and_updates(
            self,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x,
            x_generated,
    ):
        (
            y_pred_generated,
            discriminator_non_trainable_vars
        ) = self.discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x_generated)
        (
            y_pred_real,
            discriminator_non_trainable_vars
        ) = self.discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x)
        loss = self.compute_disciriminator_loss(y_pred_real, y_pred_generated)
        return loss, (discriminator_non_trainable_vars,)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        ) = state

        seed = non_trainable_variables[0]

        # Get generator state
        # Since generator comes and gets built before discriminator
        generator_trainable_vars = trainable_variables[
                                   :len(self.generator.trainable_variables)]
        generator_non_trainable_vars = non_trainable_variables[1:][
                                       :len(self.generator.non_trainable_variables)]
        generator_optimizer_vars = optimizer_variables[
                                   :len(self.generator_optimizer.variables)]

        # Get discriminator state
        discriminator_trainable_vars = trainable_variables[
                                       len(self.generator.trainable_variables):]
        discriminator_non_trainable_vars = non_trainable_variables[1:][
                                           len(self.generator.non_trainable_variables):]
        discriminator_optimizer_vars = optimizer_variables[
                                       len(self.generator_optimizer.variables):
                                       ]

        actual_data, dummy_y = data
        del dummy_y
        x, cond_vectors_real, cond_vectors, mask = actual_data
        batch_size = ops.shape(x)[0]

        # =========================
        # Discriminator train step
        # =========================
        seed = jax.random.split(seed, 1)[0]
        z = random.normal(shape=[batch_size, self.latent_dim],
                          seed=seed)
        input_gen = ops.concatenate([z, cond_vectors], axis=1)
        x_generated, _ = self.generator.stateless_call(
            generator_trainable_vars,
            generator_non_trainable_vars,
            input_gen)
        x_generated = ops.concatenate(
            [x_generated, cond_vectors], axis=1)
        x = ops.concatenate([x, cond_vectors_real],
                            axis=1)
        # w.r.t. gradient penalty
        disc_grad_fn_pen = jax.value_and_grad(
            self.discriminator_compute_grad_penalty_and_updates,
            has_aux=True)
        (grad_pen,
         (discriminator_non_trainable_vars,)), grads = disc_grad_fn_pen(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x,
            x_generated
        )
        (
            discriminator_trainable_vars,
            discriminator_optimizer_vars
        ) = self.discriminator_optimizer.stateless_apply(
            discriminator_optimizer_vars,
            grads,
            discriminator_trainable_vars
        )
        # w.r.t. loss
        disc_grad_fn_loss = jax.value_and_grad(
            self.discriminator_compute_loss_and_updates,
            has_aux=True)
        (
            d_loss, (discriminator_non_trainable_vars,)), grads = disc_grad_fn_loss(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            x,
            x_generated
        )
        (
            discriminator_trainable_vars,
            discriminator_optimizer_vars
        ) = self.discriminator_optimizer.stateless_apply(
            discriminator_optimizer_vars,
            grads,
            discriminator_trainable_vars
        )

        # =====================
        # Generator train step
        # =====================
        seed = jax.random.split(seed, 1)[0]
        z = random.normal(shape=[batch_size, self.latent_dim],
                          seed=seed)
        input_gen = ops.concatenate([z, cond_vectors], axis=1)
        gen_grad_fn = jax.value_and_grad(
            self.generator_compute_loss_and_updates,
            has_aux=True,
        )
        (g_loss, (generator_non_trainable_vars,)), grads = gen_grad_fn(
            generator_trainable_vars,
            generator_non_trainable_vars,
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            input_gen,
            cond_vectors,
            mask,
            self.metadata,
        )

        (
            generator_trainable_vars,
            generator_optimizer_vars
        ) = self.generator_optimizer.stateless_apply(
            generator_optimizer_vars,
            grads,
            generator_trainable_vars
        )

        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(new_metric_variables) + len(metric.variables)
                                    ]
            if metric.name == "generator_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    g_loss
                )
            elif metric.name == "discriminator_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    d_loss
                )
            else:
                continue
            logs[metric.name] = metric.stateless_result(this_metric_variables)
            new_metric_variables += this_metric_variables

        # Update seed
        seed = jax.random.split(seed, 1)[0]
        state = (
            generator_trainable_vars + discriminator_trainable_vars,
            [seed] + generator_non_trainable_vars +
            discriminator_non_trainable_vars,
            generator_optimizer_vars + discriminator_optimizer_vars,
            new_metric_variables
        )
        return logs, state



import keras


class Pretrainer(keras.Model):
    """Base Pretrainer class for the architectures that make use of
    semi-supervised pretraining.

    Args:
        base_model: keras.Model instance, that can be a base model or
            part of the architecture like an encoder in an encoder
            decoder model where only the encoder needs to be pretrained.
    """
    def __init__(self,
                 base_model,
                 **kwargs):
        super().__init__(**kwargs)
        self.base_model = base_model
        self._pretrained = False
        self._pretrained_model = None

    @property
    def pretrained_model(self):
        return self._pretrained_model

    @pretrained_model.setter
    def pretrained_model(self, value):
        if not isinstance(value, keras.Model):
            raise ValueError(
                "Pretrained model must be an instance of `keras.Model`. "
                f"Received: {type(value)}.")
        self._pretrained_model = value

    @pretrained_model.getter
    def pretrained_model(self):
        if not self._pretrained:
            raise Exception("Model has not yet been pretrained.")
        return self._pretrained_model

    def is_pretrained(self):
        return self._pretrained

    def get_config(self):
        config = super().get_config()
        config.update({
            "base_model": keras.layers.serialize(self.base_model),
        }
        )

    @classmethod
    def from_config(cls, config):
        config["base_model"] = keras.layers.deserialize(
                                                config.pop("base_model"))
        return cls(**config)



import keras
from keras import ops
from teras._src.layers.tabnet.feature_transformer_layer import TabNetFeatureTransformerLayer
from teras._src.api_export import teras_export
from teras._src.layers.layer_list import LayerList


@teras_export("teras.layers.TabNetFeatureTransformer")
class TabNetFeatureTransformer(keras.layers.Layer):
    """
    FeatureTransformer layer based on the TabNet architecture proposed
    in the "TabNet: Attentive Interpretable Tabular Learning" paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        hidden_dim: int, hidden dimensionality of the feature transformer.
        num_shared_layers: int, number of shared layers to use.
            Defaults to 2.
        num_decision_dependent_layers: int, number of decision dependent
            layers to use. Defaults to 2.
        batch_momentum: float, batch momentum
    """
    shared_layers = None

    def __init__(self,
                 hidden_dim: int,
                 num_shared_layers: int = 2,
                 num_decision_dependent_layers: int = 2,
                 batch_momentum: float = 0.99,
                 **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.batch_momentum = batch_momentum

        self._maybe_create_shared_layers(self.num_shared_layers,
                                         self.hidden_dim,
                                         self.batch_momentum)
        self.decision_dependent_layers = LayerList([
            TabNetFeatureTransformerLayer(
                dim=hidden_dim,
                batch_momentum=batch_momentum,
                name=f"decision_dependent_layer_{i+1}")
            for i in range(num_decision_dependent_layers)
        ],
            name="decision_dependent_layers"
        )

    @classmethod
    def _maybe_create_shared_layers(cls, num_layers, hidden_dim,
                                    batch_momentum):
        if cls.shared_layers:
            return
        cls.shared_layers = LayerList([TabNetFeatureTransformerLayer(
            dim=hidden_dim,
            batch_momentum=batch_momentum,
            name=f"shared_layer_{i+1}"
        ) for i in range(num_layers)],
            name="shared_layers"
        )

    @classmethod
    def reset_shared_layers(cls):
        cls.shared_layers = None

    def build(self, input_shape):
        # Shared layers need only be built once, because they are
        # shared across instances of `TabNetFeatureTransformer`
        if not self.shared_layers.built:
            self.shared_layers.build(input_shape)
        input_shape = self.shared_layers.compute_output_shape(input_shape)
        self.decision_dependent_layers.build(input_shape)
        self.built = True

    def call(self, inputs):
        x = inputs
        residue = None
        for layer in self.shared_layers:
            x = layer(x)
            if residue is not None:
                x += ops.sqrt(0.5) * residue
            residue = x

        for layer in self.decision_dependent_layers:
            x = layer(x)
            if residue is not None:
                x += ops.sqrt(0.5) * residue
            residue = x
        return x

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.hidden_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "hidden_dim": self.hidden_dim,
            "num_shared_layers": self.num_shared_layers,
            "num_decision_dependent_layers":
                self.decision_dependent_layers,
            "batch_momentum": self.batch_momentum,
        })



import jax
import keras
from keras import ops, random
from teras._src.backend.common.models.pretrainers.tabnet import BaseTabNetPretrainer


class TabNetPretrainer(BaseTabNetPretrainer):
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 missing_feature_probability: float = 0.3,
                 **kwargs):
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            missing_feature_probability=missing_feature_probability,
            **kwargs)

    def compute_loss_and_updates(self,
                                 trainable_variables,
                                 non_trainable_variables,
                                 x,
                                 mask,
                                 training=False):
        reconstructed, non_trainable_variables = self.stateless_call(
            trainable_variables,
            non_trainable_variables,
            x,
            mask=mask,
            training=training,
        )
        loss = self.compute_loss(x=x,
                                 x_reconstructed=reconstructed,
                                 mask=mask)
        return loss, (reconstructed, non_trainable_variables)

    def train_step(self, state, data):
        (trainable_variables,
         non_trainable_variables,
         optimizer_variables,
         metrics_variables
         ) = state
        seed = non_trainable_variables[0]
        # Sample mask
        mask = random.binomial(
            shape=ops.shape(data),
            counts=1,
            probabilities=self.missing_feature_probability,
            seed=seed,
        )
        # Grad fn
        grad_fn = jax.value_and_grad(self.compute_loss_and_updates,
                                     has_aux=True)

        # Compute gradients
        (loss, (reconstructed, non_trainable_variables)), grads = grad_fn(
            trainable_variables,
            non_trainable_variables,
            data,
            mask,
            training=True
        )

        # Update weights and optimizer vars
        (
            trainable_variables,
            optimizer_variables
        ) = self.optimizer.stateless_apply(
            optimizer_variables,
            grads,
            trainable_variables
        )
        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(new_metric_variables) + len(metric.variables)
                                    ]
            if metric.name == "loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    loss
                )
            else:
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    data,
                    reconstructed
                )
            logs[metric.name] = metric.stateless_result(this_metric_variables)
            new_metric_variables += this_metric_variables

        seed = jax.random.split(seed, 1)[0]
        non_trainable_variables[0] = seed
        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            new_metric_variables
        )
        return logs, state



import jax
import jax.numpy as jnp
import keras
from teras._src.backend.common.models.pretrainers.tab_transformer import (
    BaseTabTransformerMLMPretrainer,
    BaseTabTransformerRTDPretrainer
)
from keras import ops


class TabTransformerMLMPretrainer(BaseTabTransformerMLMPretrainer):
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 missing_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         data_dim=data_dim,
                         missing_rate=missing_rate,
                         mask_seed=mask_seed,
                         **kwargs)

    def _create_mask(self, input_shape):
        if len(input_shape) != 2:
            raise ValueError(
                "Expected `input_shape` to have a rank of 2. "
                f"Received, {input_shape} with rank {len(input_shape)}")
        batch_size, _ = input_shape
        key = jax.random.PRNGKey(self.mask_seed)
        num_features_to_miss = int(self.missing_rate * self.data_dim)
        mask = jnp.ones((num_features_to_miss,), dtype=jnp.int32)
        mask = jnp.pad(mask,
                       (0, self.data_dim - num_features_to_miss))
        mask = jnp.tile(jnp.expand_dims(mask, 0), (batch_size, 1))
        mask = jax.random.permutation(key, mask, axis=1, independent=True)
        return mask

    def compute_loss_and_updates(self,
                                 trainable_variables,
                                 non_trainable_variables,
                                 x,
                                 mask,
                                 training=False):
        y_pred, non_trainable_variables = self.stateless_call(
            trainable_variables,
            non_trainable_variables,
            x,
            mask=(1 - mask),
            training=training
        )
        loss = self.compute_loss(y=x * mask,
                                 y_pred=y_pred)
        return loss, (y_pred, non_trainable_variables)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        ) = state
        mask = self._create_mask(ops.shape(data))
        # In mask, 1 indicates that the feature will be missing, while 0
        # indicates the opposite
        grad_fn = jax.value_and_grad(self.compute_loss_and_updates,
                                     has_aux=True)
        (loss, (y_pred, non_trainable_variables)), grads = grad_fn(
            trainable_variables,
            non_trainable_variables,
            data,
            mask=mask,
            training=True
        )
        (
            trainable_variables,
            optimizer_variables
        ) = self.optimizer.stateless_apply(
            optimizer_variables,
            grads,
            trainable_variables
        )

        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(
                                        new_metric_variables) + len(
                                        metric.variables)
                                    ]
            if metric.name == "loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    loss
                )
            else:
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    data * mask,
                    y_pred
                )
            logs[metric.name] = metric.stateless_result(
                this_metric_variables)
            new_metric_variables += this_metric_variables

        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            new_metric_variables
        )
        return logs, state


class TabTransformerRTDPretrainer(BaseTabTransformerRTDPretrainer):
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 replace_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         data_dim=data_dim,
                         replace_rate=replace_rate,
                         mask_seed=mask_seed,
                         **kwargs)

    def _create_mask(self, input_shape):
        if len(input_shape) != 2:
            raise ValueError(
                "Expected `input_shape` to have a rank of 2. "
                f"Received, {input_shape} with rank {len(input_shape)}")
        batch_size, _ = input_shape
        key = jax.random.PRNGKey(self.mask_seed)
        num_features_to_replace = int(self.replace_rate * self.data_dim)
        mask = jnp.ones((num_features_to_replace,), dtype=jnp.int32)
        mask = jnp.pad(mask,
                       (0, self.data_dim - num_features_to_replace))
        mask = jnp.tile(jnp.expand_dims(mask, 0), (batch_size, 1))
        mask = jax.random.permutation(key, mask, axis=1, independent=True)
        return mask

    def compute_loss_and_updates(self,
                                 trainable_variables,
                                 non_trainable_variables,
                                 x,
                                 mask,
                                 training=False):
        y_pred, non_trainable_variables = self.stateless_call(
            trainable_variables,
            non_trainable_variables,
            x,
            mask=(1 - mask),
            training=training
        )
        loss = self.compute_loss(y=mask,
                                 y_pred=y_pred)
        return loss, (y_pred, non_trainable_variables)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        ) = state

        mask = self._create_mask(ops.shape(data))
        # In mask, 1 indicates that the feature will be missing, while 0
        # indicates the opposite
        grad_fn = jax.value_and_grad(self.compute_loss_and_updates,
                                     has_aux=True)
        (loss, (y_pred, non_trainable_variables)), grads = grad_fn(
            trainable_variables,
            non_trainable_variables,
            data,
            mask=mask,
            training=True
        )
        (
            trainable_variables,
            optimizer_variables
        ) = self.optimizer.stateless_apply(
            optimizer_variables,
            grads,
            trainable_variables
        )

        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(
                                        new_metric_variables) + len(
                                        metric.variables)
                                    ]
            if metric.name == "loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    loss
                )
            else:
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    data * mask,
                    y_pred
                )
            logs[metric.name] = metric.stateless_result(
                this_metric_variables)
            new_metric_variables += this_metric_variables

        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            new_metric_variables
        )
        return logs, state



import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase
from teras._src.models.pretrainers.tabnet.decoder import TabNetDecoder

from teras._src.models.backbones.tabnet.encoder import TabNetEncoderBackbone
from teras._src.models.pretrainers.tabnet.tabnet import TabNetPretrainer


class TabNetPretrainerTest(TestCase):
    def setUp(self):
        self.batch_size = 16
        self.input_batch = random.normal((self.batch_size, 5))
        self.mask = random.binomial((self.batch_size, 5),
                                    1, 0.5)

    def test_valid_call(self):
        encoder = TabNetEncoderBackbone(
            input_dim=5,
            feature_transformer_dim=16,
            decision_step_dim=8)
        decoder = TabNetDecoder(data_dim=5,
                                feature_transformer_dim=16,
                                decision_step_dim=8)
        pretrainer = TabNetPretrainer(
            encoder=encoder,
            decoder=decoder
        )
        reconstructed_features = pretrainer(self.input_batch,
                                            self.mask)

    def test_valid_output_shape(self):
        encoder = TabNetEncoderBackbone(
            input_dim=5,
            feature_transformer_dim=16,
            decision_step_dim=16)
        decoder = TabNetDecoder(data_dim=5,
                                feature_transformer_dim=16,
                                decision_step_dim=16)
        pretrainer = TabNetPretrainer(
            encoder=encoder,
            decoder=decoder
        )
        reconstructed_features = pretrainer(self.input_batch,
                                            self.mask)
        self.assertEqual(ops.shape(reconstructed_features),
                         (self.batch_size, 5))

    def test_fit(self):
        encoder = TabNetEncoderBackbone(
            input_dim=5,
            feature_transformer_dim=16,
            decision_step_dim=16)
        decoder = TabNetDecoder(data_dim=5,
                                feature_transformer_dim=16,
                                decision_step_dim=16)
        pretrainer = TabNetPretrainer(
            encoder=encoder,
            decoder=decoder
        )
        pretrainer.compile(optimizer=keras.optimizers.Adam())
        pretrainer.build(ops.shape(self.input_batch))
        pretrainer.fit(self.input_batch)



from keras.src.testing.test_case import TestCase
from keras import ops, random
from teras._src.layers.transformer.encoder_layer import TransformerEncoderLayer


class TransformerEncoderLayerTest(TestCase):
    def test_valid_output_shape(self):
        input_shape = (4, 5, 16)
        inputs = random.normal(shape=input_shape)
        encoder_layer = TransformerEncoderLayer(embedding_dim=16,
                                                num_heads=8,
                                                feedforward_dim=64,
                                                )
        outputs = encoder_layer(inputs)
        self.assertEqual(ops.shape(outputs), input_shape)



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.ContinuousExtraction")
class ContinuousExtraction(keras.layers.Layer):
    """
    ContinuousExtraction layer extracts continuous features from inputs
    as is. It helps us build functional model where inputs to the
    model contain both categorical and continuous features, but they
    must diverge into two different branches for separate processing.

    Args:
        continuous_idx: list, a list of indices of continuous
            features in the given dataset.
    """
    def __init__(self,
                 continuous_idx: list,
                 **kwargs):
        super().__init__(trainable=False, **kwargs)
        self.continuous_idx = continuous_idx

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape)
        output_shape[1] = len(self.continuous_idx)
        return tuple(output_shape)

    def call(self, inputs):
        continuous_features = ops.take(inputs,
                                       indices=self.continuous_idx,
                                       axis=1)
        return continuous_features

    def get_config(self):
        config = super().get_config()
        config.update({
            "continuous_idx": self.continuous_idx
        })



import os

import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase

from teras._src.models.autoencoders.tvae.decoder import TVAEDecoder


class TVAEDecoderTest(TestCase):
    def setUp(self):
        self.batch_size = 8
        self.data_dim = 5
        self.inputs = random.normal((self.batch_size, 8))

    def test_valid_call(self):
        decoder = TVAEDecoder(data_dim=self.data_dim)
        outputs = decoder(self.inputs)

    def test_valid_output_shape(self):
        decoder = TVAEDecoder(data_dim=self.data_dim)
        outputs = decoder(self.inputs)
        # Decoder returns generated samples and sigmas
        self.assertEqual(len(outputs), 2)
        # Generated samples shape check
        self.assertEqual(ops.shape(outputs[0]),
                         (self.batch_size, self.data_dim))
        # sigmas shape check
        self.assertEqual(ops.shape(outputs[1]),
                         (self.data_dim,))

    def test_model_save_and_load(self):
        decoder = TVAEDecoder(data_dim=self.data_dim)
        outputs = decoder(self.inputs)
        save_path = os.path.join(self.get_temp_dir(),
                                 "tvae_decoder.keras"
                                 )
        decoder.save(save_path)
        reloaded_model = keras.models.load_model(save_path)

        # Check we got the real object back
        self.assertIsInstance(reloaded_model, TVAEDecoder)

        # Check that output matches
        reloaded_outputs = reloaded_model(self.inputs)
        for original, reloaded in zip(outputs, reloaded_outputs):
            self.assertAllClose(
                ops.convert_to_numpy(original),
                ops.convert_to_numpy(reloaded)
            )



from keras import ops
import pandas as pd
import numpy as np
from warnings import warn
from teras._src.typing import DataFrameOrNdArray
from teras._src.api_export import teras_export


@teras_export("teras.utils.compute_cardinalities")
def compute_cardinalities(x, categorical_idx):
    """
    Compute cardinalities for features in the given dataset/dataframe.
    For numerical features, 0 is used as a placeholder.

    Args:
        x: Input dataset or dataframe.
        categorical_idx: list, a list of indices of categorical features
            in the given dataset.

    Returns:
        A 1d numpy array of cardinalities of all features.
        For numerical features, a value of 0 is used.
    """
    if isinstance(x, pd.DataFrame):
        x = x.values

    cardinalities = np.array([], dtype=np.uint16)
    for idx in range(ops.shape(x)[1]):
        if idx in categorical_idx:
            feature = ops.convert_to_numpy(x[:, idx])
            cardinalities = np.append(cardinalities,
                                      (len(np.unique(feature))))
        else:
            # it's a numerical feature, in which case we append 0
            cardinalities = np.append(cardinalities, 0)
    return cardinalities


@teras_export("teras.utils.get_metadata_for_embedding")
def get_metadata_for_embedding(dataframe: pd.DataFrame,
                               categorical_features=None,
                               numerical_features=None):
    # TODO:
    #   Add support for TensorFlow datasets and PyTorch DataLoaders/Datasets
    """
    Utility function that create metadata for features in a given dataframe
    required by the Categorical and Numerical embedding layers in Teras.
    For numerical features, it maps each feature name to feature index.
    For categorical features, it maps each feature name to a tuple of
    feature index and vocabulary of words in that categorical feature.
    This metadata is usually required by the architectures that create
    embeddings of Numerical or Categorical features,
    such as TabTransformer, TabNet, FT-Transformer, etc.

    Args:
        dataframe: Input dataframe
        categorical_features: List of names of categorical features in the
            input dataset
        numerical_features: List of names of categorical features in the
            input dataset

    Returns:
        A dictionary which contains sub-dictionaries for categorical and
        numerical features where categorical dictionary is a mapping of
        categorical feature names to a tuple of feature indices and the
        lists of unique values (vocabulary) in them, while numerical
        dictionary is a mapping of numerical feature names to their indices
        {feature_name: (feature_idx, vocabulary)} for feature in categorical features.
        {feature_name: feature_idx} for feature in numerical features.
    """
    if categorical_features is None and numerical_features is None:
        raise ValueError(
            "Both `categorical_features` and `numerical_features` cannot "
            "be None at the same time. "
            "You must pass value for at least one of them. "
            f"Received, `categorical_features`: {categorical_features}, "
            f"`numerical_features`: {numerical_features}")
    categorical_features = [] if categorical_features is None else categorical_features
    numerical_features = [] if numerical_features is None else numerical_features
    features_meta_data = {}
    categorical_features_metadata = {}
    numerical_features_metadata = {}
    # Verify all specified features are present in the dataframe
    specified_columns = set(numerical_features).union(set(categorical_features))
    not_found_in_dataframe = specified_columns - set(dataframe.columns)
    if len(not_found_in_dataframe) > 0:
        raise ValueError(
            f"Following specified features not found in the dataframe, "
            f"{not_found_in_dataframe}")
    for idx, col in enumerate(dataframe.columns):
        if categorical_features is not None and col in categorical_features:
            vocabulary = sorted(list(dataframe[col].unique()))
            categorical_features_metadata.update({col: (idx, vocabulary)})
        elif numerical_features is not None and col in numerical_features:
            numerical_features_metadata.update({col: idx})

    features_meta_data["categorical"] = categorical_features_metadata
    features_meta_data["numerical"] = numerical_features_metadata
    return features_meta_data


@teras_export("teras.utils.convert_tf_dict_to_array_tensor")
def convert_tf_dict_to_array_tensor(dict_tensor):
    """
    Converts a batch of data taken from tensorflow dictionary format
    dataset to array format.
    Args:
        dict_tensor: A batch of data taken from tensorflow dictionary
        format dataset.

    Returns:
        Array format data.
    """
    if not isinstance(dict_tensor, dict):
        warn("Given tensor is not in dictionary format."
             "Hence no processing will be applied. \n"
             f"Expected type: {dict}, Received type: {type(dict_tensor)}")
        return

    feature_names = dict_tensor.keys()
    array_tensor = [ops.expand_dims(dict_tensor[feature_name], axis=1)
                    for feature_name in feature_names]
    array_tensor = ops.concatenate(array_tensor, axis=1)
    return array_tensor


@teras_export("teras.utils.inject_missing_values")
def inject_missing_values(x: DataFrameOrNdArray,
                          miss_rate=0.1
                          ):
    """
    Injects missing (np.nan) values in the given dataframe or ndarray.

    Args:
        x: A pandas dataframe or ndarray.
        miss_rate: The fraction of missing values that should be introduced.
            Should be between 0-1. Defaults to 0.1

    Returns:
        Data with missing values.

    Example:
        ```python
        data = np.arange(1000).reshape(50, 20)
        data = inject_missing_values(data, miss_rate=0.2, return_dataframe=False)
        ```
    """
    x_with_missing_data = x.copy()
    is_dataframe = isinstance(x_with_missing_data, pd.DataFrame)

    if is_dataframe:
        x_with_missing_data = x_with_missing_data.values

    mask = np.random.binomial(1, 1-miss_rate, size=x.shape)
    x_with_missing_data[mask == 0] = np.nan

    if is_dataframe:
        x_with_missing_data = pd.DataFrame(x_with_missing_data,
                                           columns=x.columns)
    return x_with_missing_data


@teras_export("teras.utils.inject_missing_values")
def generate_fake_gemstone_data(num_samples: int = 16):
    """
    Generate fake gemstone like data of specified num_samples.

    Args:
        num_samples:
            Number of samples to generate

    Returns:
        A pandas DataFrame of fake gemstone like data.
    """
    fake_gem_df = pd.DataFrame({
        "cut": np.random.randint(low=0, high=3, size=(num_samples,)),
        "color": np.random.randint(low=0, high=5, size=(num_samples,)),
        "clarity": np.random.randint(low=0, high=4, size=(num_samples,)),
        "depth": np.random.randint(low=0, high=100, size=(num_samples,)),
        "table": np.random.randint(low=0, high=100, size=(num_samples,))
    })
    fake_gem_df = fake_gem_df.astype(np.float32)
    return fake_gem_df


@teras_export("teras.utils.clean_reloaded_config_data")
def clean_reloaded_config_data(x):
    """
    Cleans reloaded dictionary/list config data in the `from_config` method.

    Args:
        x: dict or list to clean.
    """
    if not isinstance(x, (dict, list)):
        return x
    if isinstance(x, dict):
        if "config" in x.keys():
            return x["config"]["value"]
        for key, value in x.items():
            x[key] = clean_reloaded_config_data(value)
        return x
    if isinstance(x, list):
        for i, value in enumerate(x):
            x[i] = clean_reloaded_config_data(value)
        return x



# Base task class
import keras


class Task(keras.Model):
    """
    Base class for building Task models.

    Taken from keras-cv, with some truncation.
    https://github.com/keras-team/keras-cv/blob/master/keras_cv/models/task.py
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._backbone = None

    @property
    def backbone(self):
        return self._backbone

    @backbone.setter
    def backbone(self, value):
        self._backbone = value

    def get_config(self):
        return {"name": self.name,
                "trainable": self.trainable}

    @classmethod
    def from_config(cls, config):
        if "backbone" in config and isinstance(config["backbone"], dict):
            config["backbone"] = keras.layers.deserialize(config["backbone"])
        return cls(**config)



import keras
from keras import random, ops
from teras._src.layers.ctgan.discriminator_layer import CTGANDiscriminatorLayer
from teras._src.typing import IntegerSequence
from teras._src.layers.layer_list import LayerList


class BaseCTGANDiscriminator(keras.Model):
    """
    Base CTGANDiscriminator class.
    """
    def __init__(self,
                 hidden_dims: IntegerSequence = (256, 256),
                 packing_degree: int = 8,
                 gradient_penalty_lambda: float = 10.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.hidden_dims = hidden_dims
        self.packing_degree = packing_degree
        self.gradient_penalty_lambda = gradient_penalty_lambda
        self.seed = seed
        self.seed_gen = random.SeedGenerator(self.seed)

        self.hidden_block = []
        for dim in self.hidden_dims:
            self.hidden_block.append(CTGANDiscriminatorLayer(dim))
        self.hidden_block = LayerList(
            self.hidden_block,
            sequential=True,
            name="discriminator_hidden_block"
        )
        self.output_layer = keras.layers.Dense(
            1,
            name="discriminator_output_layer")

    def build(self, input_shape):
        batch_size, input_dim = input_shape
        input_shape = (batch_size // self.packing_degree,
                       input_dim * self.packing_degree)
        self.hidden_block.build(input_shape)
        input_shape = self.hidden_block.compute_output_shape(input_shape)
        self.output_layer.build(input_shape)

    def call(self, inputs):
        inputs_dim = ops.shape(inputs)[1]
        inputs = ops.reshape(inputs,
                             newshape=(-1, self.packing_degree * inputs_dim))
        outputs = self.hidden_block(inputs)
        outputs = self.output_layer(outputs)
        return outputs

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (1,)

    def get_config(self):
        config = super().get_config()
        new_config = {
            'hidden_dims': self.hidden_dims,
            'packing_degree': self.packing_degree,
            'gradient_penalty_lambda': self.gradient_penalty_lambda,
            'seed': self.seed,
        }
        config.update(new_config)
        return config




import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.PCGAIN")
class PCGAIN(backend.models.PCGAIN):
    """
    PCGAIN is a missing data imputation model based on the GAIN architecture.
    It is proposed by Yufeng Wang et al. in the paper,
    "PC-GAIN: Pseudo-label Conditional Generative Adversarial Imputation
    Networks for Incomplete Data"

    Reference(s):
        https://arxiv.org/abs/2011.07770

    Args:
        generator: keras.Model, An instance of `GAINGenerator` model or any
            customized model that can work in its place.
        discriminator: keras.Model, An instance of `GAINDiscriminator` model
            or any customized model that can work in its place.
        classifier: keras.Model, An instance of the `PCGAINClassifier`
            trained on the imputed pretraining dataset coupled with the
            pseudo-labels generated by the clustering algorithm.
        hint_rate: float, Hint rate will be used to sample binary vectors for
            `hint vectors` generation. Must be between 0. and 1.
            Hint vectors ensure that generated samples follow the underlying
            data distribution.
            Defaults to 0.9
        alpha: float, Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely,
            `generator_loss` = `cross_entropy_loss` + `alpha` * `mse_loss`
            The higher the `alpha`, the more the mse_loss will affect the
            overall generator loss.
            Defaults to 200.
        beta: float, Hyper parameter for generator loss computation that
            controls the contribution of the classifier's loss to the
            overall generator loss. Defaults to 100.
        seed: int, seed to make results of random ops deterministic.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 classifier: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 200.,
                 beta: float = 100.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         classifier=classifier,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         beta=beta,
                         seed=seed,
                         **kwargs)



import os

import keras
from keras import ops
from keras.src.testing.test_case import TestCase
from teras._src.preprocessing.data_samplers.tvae import TVAEDataSampler

from teras._src.models.autoencoders.tvae.encoder import TVAEEncoder
from teras._src.preprocessing.data_transformers.tvae import TVAEDataTransformer
from teras._src.utils import generate_fake_gemstone_data


class TVAEEncoderTest(TestCase):
    def setUp(self):
        fake_gem_df = generate_fake_gemstone_data(num_samples=16)
        cont_cols = ["depth", "table"]
        cat_cols = ["cut", "color", "clarity"]
        data_transformer = TVAEDataTransformer(continuous_features=cont_cols,
                                               categorical_features=cat_cols)
        x_transformed = data_transformer.fit_transform(fake_gem_df)
        self.batch_size = 8
        data_sampler = TVAEDataSampler(metadata=data_transformer.get_metadata(),
                                       categorical_features=cat_cols,
                                       continuous_features=cont_cols,
                                       batch_size=self.batch_size,
                                       )
        self.dataset = data_sampler.get_dataset(x_transformed=x_transformed,
                                                x_original=fake_gem_df)
        self.data_batch = next(iter(self.dataset))
        self.metadata = data_transformer.get_metadata()
        self.data_dim = data_sampler.data_dim
        self.latent_dim = 16

    def test_valid_call(self):
        encoder = TVAEEncoder(latent_dim=self.latent_dim)
        outputs = encoder(self.data_batch)

    def test_valid_output_shape(self):
        encoder = TVAEEncoder(latent_dim=self.latent_dim)
        outputs = encoder(self.data_batch)
        # encoder returns mean, log_var
        self.assertEqual(len(outputs), 2)
        # mean output shape check
        self.assertEqual(ops.shape(outputs[0]),
                         (self.batch_size, self.latent_dim))
        # log_var output shape check (yes both have the same shape)
        self.assertEqual(ops.shape(outputs[1]),
                         (self.batch_size, self.latent_dim))

    def test_model_save_and_load(self):
        encoder = TVAEEncoder(latent_dim=self.latent_dim)
        outputs = encoder(self.data_batch)
        save_path = os.path.join(self.get_temp_dir(),
                                 "tvae_encoder.keras"
                                 )
        encoder.save(save_path)
        reloaded_model = keras.models.load_model(save_path)

        # Check we got the real object back
        self.assertIsInstance(reloaded_model, TVAEEncoder)

        # Check that output matches
        reloaded_outputs = reloaded_model(self.data_batch)
        for original, reloaded in zip(outputs, reloaded_outputs):
            self.assertAllClose(
                ops.convert_to_numpy(original),
                ops.convert_to_numpy(reloaded)
            )



import torch
import keras
from keras import random, ops
from teras._src.backend.common.models.gans.ctgan.ctgan import BaseCTGAN


class CTGAN(BaseCTGAN):
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 metadata: dict,
                 latent_dim: int = 128,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         metadata=metadata,
                         latent_dim=latent_dim,
                         seed=seed,
                         **kwargs)

    def train_step(self, data):
        x, cond_vectors_real, cond_vectors, mask = data
        batch_size = ops.shape(x)[0]

        # =========================
        # Discriminator train step
        # =========================
        self.zero_grad()
        z = random.normal(shape=[batch_size, self.latent_dim],
                          seed=self._seed_gen)
        input_gen = ops.concatenate([z, cond_vectors], axis=1)
        x_generated = self.generator(input_gen, training=False)
        x_generated = ops.concatenate(
            [x_generated, cond_vectors], axis=1)
        x = ops.concatenate([x, cond_vectors_real],
                            axis=1)

        y_pred_generated = self.discriminator(x_generated)
        y_pred_real = self.discriminator(x)
        grad_pen = self.discriminator.gradient_penalty(x,
                                                       x_generated)
        loss_disc = self.discriminator_loss(y_pred_real, y_pred_generated)
        grad_pen.backward(retain_graph=True)
        grads = [v.value.grad for v in self.discriminator.trainable_variables]
        with torch.no_grad():
            self.discriminator_optimizer.apply(
                grads,
                self.discriminator.trainable_variables)
        loss_disc.backward()
        grads = [v.value.grad for v in self.discriminator.trainable_variables]
        with torch.no_grad():
            self.discriminator_optimizer.apply(
                grads,
                self.discriminator.trainable_variables)

        # =====================
        # Generator train step
        # =====================
        self.zero_grad()
        z = random.normal(shape=[batch_size, self.latent_dim],
                          seed=self._seed_gen)
        input_gen = ops.concatenate([z, cond_vectors], axis=1)
        x_generated = self.generator(input_gen)
        x_generated = ops.concatenate(
            [x_generated, cond_vectors], axis=1)
        y_pred_generated = self.discriminator(x_generated)
        loss_gen = self.generator_loss(x_generated, y_pred_generated,
                                       cond_vectors=cond_vectors, mask=mask,
                                       metadata=self.metadata)
        loss_gen.backward()
        grads = [v.value.grad for v in self.generator.trainable_variables]
        self.generator_optimizer.apply(grads,
                                       self.generator.trainable_variables)

        self.generator_loss_tracker.update_state(loss_gen)
        self.discriminator_loss_tracker.update_state(loss_disc)

        results = {m.name: m.result() for m in self.metrics}
        return results



import keras
from keras import random, ops
from teras._src.layers.cutmix import CutMix
from teras._src.layers.mixup import MixUp
from teras._src.layers.saint.embedding import SAINTEmbedding
from teras._src.layers.saint.projection_head import SAINTProjectionHead
from teras._src.layers.saint.reconstruction_head import SAINTReconstructionHead
import numpy as np


class BaseSAINTPretrainer(keras.Model):
    def __init__(self,
                 model: keras.Model,
                 cardinalities: list,
                 embedding_dim: int,
                 cutmix_probability: float = 0.3,
                 mixup_alpha: float = 1.,
                 temperature: float = 0.7,
                 lambda_: float = 10.,
                 lambda_c: float = 0.5,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.cardinalities = cardinalities
        self.embedding_dim = embedding_dim
        self.cutmix_probability = cutmix_probability
        self.mixup_alpha = mixup_alpha
        self.temperature = temperature
        self.lambda_ = lambda_
        self.lambda_c = lambda_c
        self.seed = seed

        self.cutmix = CutMix(probability=self.cutmix_probability,
                             mask_seed=self.seed,
                             shuffle_seed=self.seed + 10)
        self.mixup = MixUp(alpha=self.mixup_alpha,
                           lambda_seed=self.seed + 20,
                           shuffle_seed=self.seed + 30)
        self.embedding = SAINTEmbedding(
            embedding_dim=self.embedding_dim,
            cardinalities=self.cardinalities)

        # For the computation of contrastive loss, we use projection heads.
        # Projection head hidden dimensions as calculated by the
        # official implementation
        data_dim = len(cardinalities)
        projection_head_hidden_dim = 6 * embedding_dim * data_dim // 5
        projection_head_output_dim = embedding_dim * data_dim // 2
        self.projection_head_real = SAINTProjectionHead(
            hidden_dim=projection_head_hidden_dim,
            output_dim=projection_head_output_dim,
            name="projection_head_real"
        )
        self.projection_head_mixed = SAINTProjectionHead(
            hidden_dim=projection_head_hidden_dim,
            output_dim=projection_head_output_dim,
            name="projection_head_mixed"
        )
        self.flatten = keras.layers.Flatten()

        self.reconstruction_head = SAINTReconstructionHead(
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
            name="reconstruction_head"
        )

        self.loss_tracker = keras.metrics.Mean(name="loss")
        self.constrastive_loss_tracker = keras.metrics.Mean(
            name="constrastive_loss"
        )
        self.denoising_loss_tracker = keras.metrics.Mean(
            name="denoising_loss"
        )

        self._pretrained = False

    def build(self, input_shape):
        self.cutmix.build(input_shape)
        self.mixup.build(input_shape)
        self.embedding.build(input_shape)
        input_shape = self.embedding.compute_output_shape(input_shape)
        self.model.build(input_shape)
        input_shape = self.model.compute_output_shape(input_shape)
        self.projection_head_real.build(input_shape)
        self.projection_head_mixed.build(input_shape)
        self.flatten.build(input_shape)
        self.reconstruction_head.build(input_shape)

    def compute_loss(self, x, x_reconstructed, z, z_augmented, cardinalities,
                     temperature=0.7, lambda_=10., lambda_c=0.5):
        """

        Args:
            x: Samples drawn from the original dataset.
            x_reconstructed: Samples reconstructed by the reconstruction head.
            z: Encodings of the real samples
            z_augmented: Encodings of the augmented samples
            cardinalities: list, a list cardinalities of all the features
                in the dataset in the same order as the features' occurrence.
                For numerical features, use the value `0` as indicator at
                the corresponding index.
                You can use the `compute_cardinalities` function from
                `teras.utils` package for this purpose.
            temperature: float, used in the computation of the
                `contrastive_loss` to scale logits. Defaults to 0.7
            lambda_: float, acts as a weight when adding the contrastive loss
                and the denoising loss together.
                `loss = constrastive_loss + lambda_ * denoising_loss`
                Defaults to 10.
            lambda_c: float, used in the computation of the contrastive
                loss. Similar to `lambda_` is helps combined two sub-losses
                within the contrastive loss. Defaults to 0.5

        Returns:
            Combined contrastive and denoising loss.
        """
        # ====================
        # Constrastive loss
        # ====================
        batch_size = ops.shape(z)[0]
        labels = ops.arange(batch_size)
        logits_ab = ops.matmul(z, ops.transpose(z_augmented)) / temperature
        logits_ba = ops.matmul(z_augmented, ops.transpose(z)) / temperature
        loss_a = keras.losses.SparseCategoricalCrossentropy(from_logits=True)(
            y_true=labels,
            y_pred=logits_ab)
        loss_b = keras.losses.SparseCategoricalCrossentropy(from_logits=True)(
            y_true=labels,
            y_pred=logits_ba)
        c_loss = lambda_c * (loss_a + loss_b) / 2

        # ==================
        # Denoising loss
        # ==================
        num_features = len(cardinalities)
        categorical_features_idx = np.flatnonzero(cardinalities)
        num_categorical_features = len(categorical_features_idx)
        categorical_cardinalities = np.asarray(cardinalities)[
            categorical_features_idx]
        total_categories = sum(cardinalities)
        d_loss = 0.
        # The reconstructed samples have dimensions equal to number of
        # categorical features + number of categories in all the categorical
        # features combined.
        # In other words, each categorical feature gets expanded into
        # `number of categories` features due to softmax layer during the
        # reconstruction phase.
        # Since the order they occur in is still the same
        # except for that the first `total_categories` number of features are
        # categorical and following are continuous
        for current_idx, feature_idx in enumerate(categorical_features_idx):
            num_categories = categorical_cardinalities[current_idx]
            d_loss += keras.losses.SparseCategoricalCrossentropy(
                from_logits=True)(
                # real_samples have the original feature order
                y_true=x[:, feature_idx],
                y_pred=x_reconstructed[:, current_idx: current_idx + num_categories]
            )
            current_idx += 1
        # Check if there are numerical features -- if so, compute the mse loss
        mse_loss = ops.cond(pred=num_categorical_features < num_features,
                            true_fn=lambda: ops.sum(
                                keras.losses.MeanSquaredError()(
                                    x[:, num_categorical_features:],
                                    x_reconstructed[:, total_categories:])),
                            false_fn=lambda: 0.
                            )
        d_loss += mse_loss
        loss = c_loss + lambda_ * d_loss
        return loss, c_loss, d_loss

    def reset_metrics(self):
        self.loss_tracker.reset_state()
        self.constrastive_loss_tracker.reset_state()
        self.denoising_loss_tracker.reset_state()

    @property
    def metrics(self):
        metrics = [self.loss_tracker,
                   self.constrastive_loss_tracker,
                   self.denoising_loss_tracker,
                   ]
        return metrics

    def call(self, inputs, **kwargs):
        real = inputs
        # Apply cutmix to raw inputs
        mixed = self.cutmix(inputs)

        # Apply embedding layer to raw inputs as well as cutmixed inputs
        real = self.embedding(real)
        mixed = self.embedding(mixed)

        # Apply mixup in embedding space
        mixed = self.mixup(mixed)

        # Pass these through encoder
        real = self.model(real)
        mixed = self.model(mixed)

        # For contrastive loss
        z_real = self.projection_head_real(real)
        z_real = self.flatten(z_real)
        z_mixed = self.projection_head_mixed(mixed)
        z_mixed = self.flatten(z_mixed)

        # For denoising loss
        reconstructed = self.reconstruction_head(mixed)

        return (z_real, z_mixed), reconstructed

    @property
    def pretrained_embedding_layer(self):
        if not self._pretrained:
            raise AssertionError(
                "The `fit` method of the `SAINTPretrainer` has not yet "
                "been called. Please train the it before accessing the "
                "`pretrained_embedding_layer` attribute."
            )
        return self.embedding

    @property
    def pretrained_model(self):
        if not self._pretrained:
            raise AssertionError(
                "The `fit` method of the `TabTransformerMLMPretrainer` "
                "has not yet been called. Please train the it before "
                "accessing the `pretrained_model` attribute."
            )
        return self.model

    def get_config(self):
        config = {
            "name": self.name,
            "trainable": self.trainable,
            "model": keras.layers.serialize(self.model),
            "cardinalities": self.cardinalities,
            "cutmix_probability": self.cutmix_probability,
            "mixup_alpha": self.mixup_alpha,
            "temperature": self.temperature,
            "lambda_": self.lambda_,
            "lambda_c": self.lambda_c,
            "seed": self.seed
        }
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        return cls(model=model, **config)




import keras
from keras import random, ops
from keras.backend import floatx


class BaseTabNetPretrainer(keras.Model):
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 missing_feature_probability: float = 0.3,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.encoder = encoder
        self.decoder = decoder
        self.missing_feature_probability = missing_feature_probability
        self.seed = seed

        self.loss_tracker = keras.metrics.Mean(name="loss")
        self._seed_generator = random.SeedGenerator(seed=self.seed)
        self._pretrained = False

    def build(self, input_shape):
        self.encoder.build(input_shape)
        input_shape = self.encoder.compute_output_shape(input_shape)
        self.decoder.build(input_shape)

    def compute_loss(self, x, x_reconstructed, mask):
        nominator_part = (x_reconstructed - x) * mask
        real_samples_population_std = ops.std(ops.cast(x, dtype=floatx()))
        # divide
        x = nominator_part / real_samples_population_std
        # Calculate L2 norm
        loss = ops.sqrt(ops.sum(ops.square(x)))
        return loss

    def reset_metrics(self):
        self.loss_tracker.reset_state()

    @property
    def metrics(self):
        return [self.loss_tracker]

    def call(self, inputs, mask, **kwargs):
        x = inputs * (1 - mask)
        # Encoded representations
        x = self.encoder(x, mask=(1 - mask))
        # Reconstructed features
        x = self.decoder(x, mask=mask)
        self._pretrained = True
        return x

    def get_config(self):
        config = {
            "name": self.name,
            "trainable": self.trainable,
            "encoder": keras.layers.serialize(self.encoder),
            "decoder": keras.layers.serialize(self.decoder),
            "missing_feature_probability":
                self.missing_feature_probability,
            "seed": self.seed,
        }
        return config

    @classmethod
    def from_config(cls, config):
        encoder = keras.layers.deserialize(config.pop("encoder"))
        decoder = keras.layers.deserialize(config.pop("decoder"))
        return cls(encoder=encoder, decoder=decoder, **config)

    @property
    def pretrained_encoder(self):
        if not self._pretrained:
            raise AssertionError(
                "The `fit` method of the `TabNetPretrainer` has not yet "
                "been called. Please train the `TabNetPretrainer` before "
                "accessing the `pretrained_encoder` attribute."
            )
        return self.encoder

    @property
    def pretrained_decoder(self):
        if not self._pretrained:
            raise AssertionError(
                "The `fit` method of the `TabNetPretrainer` has not yet "
                "been called. Please train the `TabNetPretrainer` before "
                "accessing the `pretrained_decoder` attribute."
            )
        return self.encoder



import numpy as np

from teras._src.typing import FeaturesNamesType

try:
    import tensorflow as tf
except:
    raise ImportError(
        "You need tensorflow to use CTGANDataSampler. "
        "Install it using `pip install tensorflow`"
    )
import tensorflow as tf


class CTGANDataSampler:
    """
    CTGANDataSampler class based on the data sampler class
    in the official CTGAN implementation.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        metadata: dict, A dictionary of metadata computed during data
            transformation. You can access it from the ``.get_metadata()`` of
            ``CTGANDataTransformer`` instance.
        categorical_features: list, List of categorical features names.
            CTGAN requires dataset to have at least one categorical feature,
            if your dataset doesn't contain any categorical features,
            consider using some other generative model.
        continuous_features: list, List of continuous features names
        batch_size: int, default 512,
            Batch size to use for the dataset.
        seed: int, Seed for random ops.
    """
    def __init__(self,
                 metadata,
                 categorical_features: FeaturesNamesType,
                 continuous_features: FeaturesNamesType = None,
                 batch_size: int = 512,
                 seed: int = 1337):
        self.metadata = metadata
        self.batch_size = batch_size
        self.categorical_features = categorical_features
        self.continuous_features = continuous_features
        self.seed = seed

        self._np_rng = np.random.default_rng(self.seed)
        self.num_samples = None
        self.data_dim = None
        self.batch_size = batch_size
        self.row_idx_by_categories = list()
        self._num_categorical_features = len(self.categorical_features)
        self._cat_features_relative_idx = np.asarray(
            self.metadata["categorical"]["relative_indices_all"])
        self._total_categories = self.metadata["categorical"]["total_num_categories"]
        self._all_categories = self.metadata["categorical"]["categories_all"]
        self._categories_probs_all = self.metadata["categorical"]["categories_probs_all"]

        # Since the nested lists have different lengths, so lets pad
        max_num_categories = max([len(categories)
                                  for categories in self._all_categories])
        self._features_categories_probs = np.array([
            np.pad(probs, (0, max_num_categories - len(probs)),
                   constant_values=0.)
            for probs in self._categories_probs_all
        ])

    def get_dataset(self,
                    x_transformed,
                    x_original):
        """
        Args:
            x_transformed: Dataset transformed using DataTransformer class
            x_original: Original Dataset - a pandas DataFrame.
                It is used for computing categorical values' probabilities
                for later sampling.
        Returns:
            Returns a tensorflow dataset that utilizes the sample_data method
            to create batches of data. This way user can just pass the dataset
            object to the fit method of the model and each batch generated
            will satisfy all out requirements of sampling
        """
        self.num_samples, self.data_dim = x_transformed.shape
        # adapting the approach from the official implementation
        # to sample evenly across the categories to combat imbalance
        row_idx_raw = [x_original.groupby(feature).groups
                       for feature in self.categorical_features]
        self.row_idx_by_categories = [
            [values.to_list() for values in feat.values()]
            for feat in row_idx_raw]

        total_num_categories = self.metadata["categorical"]["total_num_categories"]

        dataset = tf.data.Dataset.from_generator(
            self.generator,
            output_signature=((
                tf.TensorSpec(
                    shape=(None, self.data_dim),
                    name="real_samples"),
                tf.TensorSpec(
                    shape=(None, total_num_categories),
                    dtype=tf.float32, name="cond_vectors_real"),
                tf.TensorSpec(
                    shape=(None, total_num_categories,),
                    dtype=tf.float32, name="cond_vectors"),
                tf.TensorSpec(
                    shape=(None, self._num_categorical_features,),
                    dtype=tf.float32, name="mask")
            ),
                tf.TensorSpec(shape=(None, 1),
                              dtype=tf.float32, name="dummy_vals"),
            ),
            args=(x_transformed,)
        )
        return dataset

    def sample_cond_vectors_for_training(self, batch_size):
        # 1. Create Nd zero-filled mask vectors mi = [mi(k)] where k=1...|Di|
        # and for i = 1,...,Nd, so the ith mask vector corresponds to the ith
        # column, and each component is associated to the category of that
        # column.
        mask = np.zeros((batch_size, self._num_categorical_features))
        cond_vectors = np.zeros((batch_size, self._total_categories))
        # 2. Randomly select a discrete column Di out of all the Nd discrete
        # columns, with equal probability.
        selected_cat_features_idx = self._np_rng.choice(
            np.arange(self._num_categorical_features),
            size=batch_size
        )
        selected_cat_features_relative_idx = self._cat_features_relative_idx[
            selected_cat_features_idx]

        # 3. Construct a PMF across the range of values of the column selected
        # in 2, Di* , such that the probability mass of each value is the
        # logarithm of its frequency in that column.

        # NOTE: We've precomputed the probabilities in the DataTransformer
        # class for each feature already to speed things up.
        selected_features_categories_probs = self._features_categories_probs[
            selected_cat_features_idx]

        # Choose random values idx for features
        selected_cat_values_idx = np.array([
            self._np_rng.choice(np.arange(len(probs)),
                                p=probs)
            for probs in selected_features_categories_probs]
        ).astype(np.int32)

        # Offset this index by relative index of the feature that it belongs to.
        # because the final cond vector is the concatenation of all features and
        # is just one vector that has the length equal to total_num_categories
        selected_cat_values_idx_offsetted = selected_cat_features_relative_idx

        cond_vectors[np.arange(batch_size), selected_cat_values_idx_offsetted] = 1
        mask[np.arange(batch_size), selected_cat_features_idx] = 1
        return (cond_vectors, mask, selected_cat_features_idx,
                selected_cat_values_idx)

    def sample_cond_vectors_for_generation(self, batch_size):
        """
        The difference between this method and the training one is that, here
        we sample indices purely randomly instead of based on the calculated
        probability as proposed in the paper.
        """
        num_categories_all = np.array(
            self.metadata["categorical"]["num_categories_all"])
        cond_vectors = np.zeros((batch_size, self._total_categories))
        selected_cat_features_idx = self._np_rng.choice(
            np.arange(self._num_categorical_features),
            size=batch_size
        )
        selected_cat_features_relative_idx = self._cat_features_relative_idx[
            selected_cat_features_idx]

        # For each randomly picked feature, we get it's corresponding
        # num_categories
        selected_num_categories_all = num_categories_all[selected_cat_features_idx]
        # Then we select one category index from a feature using a range of
        # 0 — num_categories
        selected_values_idx = np.array([
            self._np_rng.choice(np.arange(num_categories))
            for num_categories in selected_num_categories_all]
        ).astype(np.int32)
        # Offset this index by relative index of the feature that it belongs to.
        # because the final cond vector is the concatenation of all features and
        # is just one vector that has the length equal to total_num_categories
        selected_values_idx += selected_cat_features_relative_idx
        cond_vectors[np.arange(batch_size), selected_values_idx] = 1
        return cond_vectors

    def generator(self, x_transformed):
        """
        Used to create a tensorflow dataset.
        Returns:
            A batch of data
        """
        # This random_feature_indices variable is required during the
        # sample_cond vector method but since we're using sample_data
        # function to create out tensorflow dataset, this gets called first
        # to generate a batch, so keep in mind that this is where this
        # variable gets its values. We could alternatively just return these
        # indices and pass them as argument to the sample cond_vec but for
        # now let's just work with it.
        num_steps_per_epoch = self.num_samples // self.batch_size
        for _ in range(num_steps_per_epoch):
            # `cond_vectors` will be first concatenated with the noise
            # vector `z` to create generator input and then will be concatenated
            # with the generated samples to serve as input for discriminator
            (
                cond_vectors, mask, selected_cat_features_idx,
                selected_cat_values_idx
            ) = self.sample_cond_vectors_for_training(self.batch_size)

            # the official implementation uses actual indices during the
            # sample_cond_vector method but uses the shuffled version in
            # sampling data, so we're gonna do just that.
            shuffled_idx = np.arange(self.batch_size)
            self._np_rng.shuffle(shuffled_idx)

            shuffled_cat_features_idx = selected_cat_features_idx[shuffled_idx]
            shuffled_values_idx = selected_cat_values_idx[shuffled_idx]

            sample_idx = []
            for feat_id, val_id in zip(shuffled_cat_features_idx,
                                       shuffled_values_idx):
                s_id = self._np_rng.choice(
                    self.row_idx_by_categories[np.squeeze(feat_id)][
                        np.squeeze(val_id)])
                sample_idx.append(np.squeeze(s_id))

            # we also return shuffled_idx because it will be required to shuffle
            # the conditional vector in the training loop as we want to keep
            # the shuffling consistent as the batch of transformed data and
            # cond vector must have one to one feature correspondence.

            # `cond_vectors_real` will be concatenated with the real_samples
            # and passed to the discriminator
            cond_vectors_real = cond_vectors[shuffled_idx]
            real_samples = x_transformed[sample_idx]
            dummy_ys = np.ones((self.batch_size, 1))
            yield (real_samples, cond_vectors_real, cond_vectors, mask), dummy_ys



from keras import ops, random
from teras._src.layers.continuous_extraction import ContinuousExtraction
from keras.src.testing.test_case import TestCase


class ContinuousExtractionTest(TestCase):
    def setUp(self):
        self.input_batch = ops.array(
            [random.normal((16,)),
             ops.repeat(ops.array([1., 2., 3., 4.]), 4),
             random.normal((16,)),
             ops.repeat(ops.array([10., 20.]), 8),
             random.normal((16,))])
        self.input_batch = ops.transpose(self.input_batch)
        self.continuous_idx = [0, 2, 4]

    def test_valid_output_shape(self):
        extraction = ContinuousExtraction(
            continuous_idx=self.continuous_idx)
        continuous_features = extraction(self.input_batch)
        self.assertEqual((16, 3), ops.shape(continuous_features))



from abc import abstractmethod
import pandas as pd


class DataTransformer:
    """
    Base `DataTransformer` class.
    It provides the common methods and attributes.
    """
    def __init__(self):
        self._metadata = dict()
        self._metadata["continuous"] = dict()
        self._metadata["categorical"] = dict()
        self._fitted = False

    @abstractmethod
    def fit(self, x):
        raise NotImplementedError

    @abstractmethod
    def transform(self, x):
        raise NotImplementedError

    def get_metadata(self):
        """
        Returns:
            dictionary of features metadata.
        """
        if not self._fitted:
            raise AssertionError(
                "The `fit` method has not yet been called. Please fit the "
                "`DataTransformer` before accessing the `get_metadata` method."
            )
        return self._metadata

    @property
    def metadata(self):
        return self.get_metadata()

    def fit_transform(self,
                      x: pd.DataFrame):
        self.fit(x)
        self._fitted = True
        return self.transform(x)

    @abstractmethod
    def save(self, filename):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def load(cls, filename):
        raise NotImplementedError



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.CLSTokenExtraction")
class CLSTokenExtraction(keras.layers.Layer):
    """
    Extracts CLS Token embeddings.
    Main purpose is to make it easy to build sequential or functional
    models.

    Args:
        axis: int, defaults to 1.

    Shapes:
        Input Shape: `(batch_size, num_features, embedding_dim)`
        Output Shape: `(batch_size, 1, emebdding_dim)`
    """
    def __init__(self,
                 axis: int = 1,
                 **kwargs):
        super().__init__(**kwargs)
        self.axis = axis

    def call(self, inputs):
        return ops.take(inputs, indices=[0], axis=self.axis)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], 1, input_shape[2])

    def get_config(self):
        config = super().get_config()
        config.update({
            "axis": self.axis,
        })
        return config



import keras
from keras import ops
import numpy as np
from teras._src.api_export import teras_export


@teras_export("teras.losses.saint_constrastive_loss")
def saint_constrastive_loss(
        real,
        augmented,
        temperature: float = 0.7,
        lambda_: float = 0.5):
    """
    Info-NCE inspired contrastive loss for the pretraining objective in
    the SAINT architecture proposed in the paper,
    "SAINT: Improved Neural Networks for Tabular Data".

    Args:
        real: Encodings of the real samples
        augmented: Encodings of the augmented samples
        temperature: float, Temperature value is used in scaling the
            logits. Defaults to 0.7
        lambda_: float, determines how the losses are combined.
            Defaults to 0.5
    """
    batch_size = ops.shape(real)[0]
    labels = ops.arange(batch_size)
    logits_ab = ops.matmul(real, ops.transpose(augmented)) / temperature
    logits_ba = ops.matmul(augmented, ops.transpose(real)) / temperature
    loss_a = keras.losses.SparseCategoricalCrossentropy(from_logits=True)(
        y_true=labels,
        y_pred=logits_ab)
    loss_b = keras.losses.SparseCategoricalCrossentropy(from_logits=True)(
        y_true=labels,
        y_pred=logits_ba)
    loss = lambda_ * (loss_a + loss_b) / 2
    return loss


@teras_export("teras.losses.saint_denoising_loss")
def saint_denoising_loss(real,
                         reconstructed,
                         cardinalities: list):
    """
    Since we apply categorical and numerical embedding layers
    separately and then combine them into a new features matrix
    this effectively makes the first k features in the outputs
    categorical (since categorical embeddings are applied first)
    and all other features numerical.
    Here, k = num_categorical_features

    Args:
        real: Samples drawn from the original dataset.
        reconstructed: Samples reconstructed by the reconstruction head.
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use the value `0` as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
    """
    if len(cardinalities) != ops.shape(real)[1]:
        raise AssertionError(
            "`cardinalities` must have the length equal to the number of "
            "features in the dataset. "
            f"Received, len(cardinalities)={len(cardinalities)} and `real`"
            f" with dimensions f{ops.shape(real)[1]}"
        )
    num_features = len(cardinalities)
    categorical_features_idx = np.flatnonzero(cardinalities)
    num_categorical_features = len(categorical_features_idx)
    categorical_cardinalities = np.asarray(cardinalities)[
        categorical_features_idx]
    total_categories = sum(cardinalities)
    loss = 0.
    # The reconstructed samples have dimensions equal to number of
    # categorical features + number of categories in all the categorical
    # features combined.
    # In other words, each categorical feature gets expanded into
    # `number of categories` features due to softmax layer during the
    # reconstruction phase.
    # Since the order they occur in is still the same
    # except for that the first `total_categories` number of features are
    # categorical and following are continuous
    for current_idx, feature_idx in enumerate(categorical_features_idx):
        num_categories = categorical_cardinalities[current_idx]
        loss += keras.losses.SparseCategoricalCrossentropy(
            from_logits=True)(
            # real_samples have the original feature order
            y_true=real[:, feature_idx],
            y_pred=reconstructed[:, current_idx: current_idx + num_categories]
        )
        current_idx += 1
    # Check if there are numerical features -- if so, compute the mse loss
    mse_loss = ops.cond(pred=num_categorical_features < num_features,
                        true_fn=lambda: ops.sum(
                            keras.losses.MeanSquaredError()(
                                real[:, num_categorical_features:],
                                reconstructed[:, total_categories:])),
                        false_fn=lambda: 0.
                        )
    loss += mse_loss
    return loss



import os
os.environ["KERAS_BACKEND"] = "jax"
import keras
import tensorflow as tf
from teras._src.backend.jax.models.gans.gain import GAIN
from teras._src.models.gans.gain.generator import GAINGenerator
from teras._src.models.gans.gain.discriminator import GAINDiscriminator
import numpy as np
from teras._src.utils import inject_missing_values


DATA_DIM = 7

NUM_SAMPLES = 256
BATCH_SIZE = 32

rng = np.random.default_rng(1337)
x_gen = rng.uniform(0., high=10., size=(NUM_SAMPLES, DATA_DIM)).astype(np.float32)
x_gen = inject_missing_values(x_gen)
x_disc = rng.uniform(0., high=10., size=(NUM_SAMPLES, DATA_DIM)).astype(np.float32)
x_disc = inject_missing_values(x_disc)
input_ds = tf.data.Dataset.from_tensor_slices((x_gen, x_disc))
input_ds = input_ds.batch(BATCH_SIZE)

generator = GAINGenerator(data_dim=DATA_DIM)
discriminator = GAINDiscriminator(data_dim=DATA_DIM)
gain = GAIN(generator,
            discriminator)
gain.build(tuple(x_gen.shape))
gain.compile(generator_optimizer=keras.optimizers.Adam(),
             discriminator_optimizer=keras.optimizers.Adam())
gain.build_optimizers()
history = gain.fit(input_ds, epochs=10)



from teras._src import backend
from teras._src.typing import IntegerSequence
from teras._src.api_export import teras_export


@teras_export("teras.models.CTGANDiscriminator")
class CTGANDiscriminator(backend.models.CTGANDiscriminator):
    """
    CTGANDiscriminator for CTGAN architecture as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        hidden_dims: Sequence, A sequence of integers that is used to
            construct the hidden block.
            For each value, a `CTGANDiscriminatorLayer` of that
            dimensionality is added. Defaults to [256, 256]
        packing_degree: int, Packing degree - taken from the PacGAN paper.
            The number of samples concatenated or "packed" together.
            It must be a factor of the batch_size.
            Packing degree is borrowed from the PacGAN
            [Diederik P. Kingma et al.] architecture,
            which proposes passing `m` samples at once to discriminator
            instead of `1` to be
            jointly classified as real or fake by the discriminator in
            order to tackle the
            issue of mode collapse inherent in the GAN based architectures.
            The number of samples passed jointly `m`, is termed as the
            `packing degree`.
            Defaults to 8.
        gradient_penalty_lambda: float, Controls the strength of gradient
            penalty. lambda value is directly proportional to the strength of
            gradient penalty.
            Gradient penalty penalizes the discriminator for large
            weights in an attempt to combat Discriminator becoming
            too confident and overfitting.
            Defaults to 10.
        seed: int, Seed for random sampling. Defaults to 1337.
    """
    def __init__(self,
                 hidden_dims: IntegerSequence = (256, 256),
                 packing_degree: int = 8,
                 gradient_penalty_lambda: float = 10.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(hidden_dims=hidden_dims,
                         packing_degree=packing_degree,
                         gradient_penalty_lambda=gradient_penalty_lambda,
                         seed=seed,
                         **kwargs)


import keras
from keras import random, ops
from teras._src.models.gans.ctgan.ctgan import CTGAN
from teras._src.models.gans.ctgan.generator import CTGANGenerator
from teras._src.api_export import teras_export


@teras_export("teras.tasks.Generator")
class Generator:
    """
    Generator class that provides methods related to data generation.

    Args:
        model: keras.Model, instance of the trained model that will be
            used to generate data
        data_transformer: Instance of data transformer used to transform data
            for training.
        data_sampler: Instance of the data sampler used to sample data for
            training.
    """
    def __init__(self,
                 model: keras.Model,
                 data_transformer,
                 data_sampler=None,
                 ):
        self.model = model
        self.data_transformer = data_transformer
        self.data_sampler = data_sampler

    def generate(self, num_samples, latent_dim, batch_size=None,
                 verbose="auto", steps=None, callbacks=None, seed=None):
        """
        Generates new data samples.
        It exposes all the arguments taken by the `predict` method.

        Args:
            num_samples: int, number of samples to generate.
            latent_dim: int, latent dimensions for sampling noise. It should
                be the same as used in the model during training.
        """
        z = random.normal((num_samples, latent_dim), seed=seed)
        if isinstance(self.model, (CTGAN, CTGANGenerator)):
            if self.data_sampler is None:
                raise ValueError(
                    "For `CTGAN` architecture `data_sampler` cannot be `None`."
                    "you must pass the data sampler instance that was used to "
                    "train the architecture. "
                    f"Received: {self.data_sampler}"
                )
            cond_vectors = self.data_sampler.sample_cond_vectors_for_generation(
                batch_size=num_samples
            )
            z = ops.concatenate([z, cond_vectors], axis=1)

        x_generated = self.model.predict(z,
                                         batch_size=batch_size,
                                         verbose=verbose,
                                         steps=steps,
                                         callbacks=callbacks
                                         )
        x_generated = self.data_transformer.reverse_transform(x_generated)
        return x_generated




import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase

from teras._src.models.backbones.saint.saint import SAINTBackbone
from teras._src.models.pretrainers.saint import SAINTPretrainer


class SAINTPretrainerTest(TestCase):
    def setUp(self):
        # we assume there are 4 categorical features in the dataset,
        # with feature 1 having 5 unique categories, feature 2 having 3
        # unique categories, feature 3 with 2 unique categories and
        # feature 4 with 7 unique categories
        self.batch_size = 16
        x_cat = ops.array([
            random.randint((self.batch_size,), minval=0, maxval=card)
            for card in [5, 3, 2, 7]],
            dtype="float32")
        x_cat = ops.transpose(x_cat)
        x_cont = random.normal((self.batch_size, 3), dtype="float32")
        self.input_batch = ops.concatenate([x_cat, x_cont], axis=1)
        self.cardinalities = [5, 3, 2, 7, 0, 0, 0]  # 0 for continuous
        self.num_categorical = 4    # Number of categorical features
        self.num_continuous = 3     # Number of continuous features
        self.input_dim = 7
        self.embedding_dim = 16

    def test_valid_call(self):
        model = SAINTBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
            embedd_inputs=False,
            return_cls_token_only=False,
        )
        pretrainer = SAINTPretrainer(
            model=model,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        outputs = pretrainer(self.input_batch)

    def test_valid_output_shape(self):
        model = SAINTBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
            embedd_inputs=False,
            return_cls_token_only=False,
        )
        pretrainer = SAINTPretrainer(
            model=model,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        # Expected: outputs = (z_real, z_prime), reconstructed
        outputs = pretrainer(self.input_batch)
        self.assertEqual(len(outputs), 2)
        self.assertEqual(len(outputs[0]), 2)

    def test_valid_encodings_shape(self):
        model = SAINTBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
            embedd_inputs=False,
            return_cls_token_only=False,
        )
        pretrainer = SAINTPretrainer(
            model=model,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
        )
        outputs = pretrainer(self.input_batch)
        (z_real, z_mixed), reconstructed = outputs
        expected_z_shape = (
            self.batch_size,
            # +1 for the CLS feature
            (self.embedding_dim * self.input_dim // 2) * (self.input_dim
                                                          + 1)
        )
        self.assertEqual(ops.shape(z_real), expected_z_shape)
        self.assertEqual(ops.shape(z_mixed), expected_z_shape)

    def test_valid_reconstructed_features_shape(self):
        model = SAINTBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
            embedd_inputs=False,
            return_cls_token_only=False,
        )
        pretrainer = SAINTPretrainer(
            model=model,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        outputs = pretrainer(self.input_batch)
        (z_real, z_mixed), reconstructed = outputs
        expected_reconstructed_shape = (
            self.batch_size,
            sum(self.cardinalities) + self.num_continuous)
        self.assertEqual(ops.shape(reconstructed),
                         expected_reconstructed_shape)

    def test_fit(self):
        model = SAINTBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim,
            embedd_inputs=False,
            return_cls_token_only=False,
        )
        pretrainer = SAINTPretrainer(
            model=model,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer.compile(optimizer=keras.optimizers.Adam())
        pretrainer.build(ops.shape(self.input_batch))
        pretrainer.fit(self.input_batch)



import keras
from keras.src.testing.test_case import TestCase
from teras._src.preprocessing.data_samplers.ctgan import CTGANDataSampler
from teras._src.preprocessing.data_transformers.ctgan import CTGANDataTransformer

from teras._src.models.gans.ctgan.ctgan import CTGAN
from teras._src.models.gans.ctgan.discriminator import CTGANDiscriminator
from teras._src.models.gans.ctgan.generator import CTGANGenerator
from teras._src.utils import generate_fake_gemstone_data


class CTGANTest(TestCase):
    def setUp(self):
        fake_gem_df = generate_fake_gemstone_data(num_samples=256)
        cont_cols = ["depth", "table"]
        cat_cols = ["cut", "color", "clarity"]
        data_transformer = CTGANDataTransformer(continuous_features=cont_cols,
                                                categorical_features=cat_cols)
        x_transformed = data_transformer.fit_transform(fake_gem_df)

        data_sampler = CTGANDataSampler(metadata=data_transformer.get_metadata(),
                                        categorical_features=cat_cols,
                                        continuous_features=cont_cols,
                                        batch_size=32,
                                        )
        self.dataset = data_sampler.get_dataset(x_transformed=x_transformed,
                                                x_original=fake_gem_df)
        self.metadata = data_transformer.get_metadata()
        self.data_dim = data_sampler.data_dim
        # if keras.backend.backend() == "torch":
        #     from torch.utils.data import DataLoader
        #     self.input_ds = DataLoader(self.input_ds, batch_size=8)
        # else:
        #     self.input_ds = self.input_ds.batch(batch_size=8)

    def test_fit(self):
        generator = CTGANGenerator(data_dim=self.data_dim,
                                   metadata=self.metadata)
        discriminator = CTGANDiscriminator()
        ctgan = CTGAN(generator=generator,
                      discriminator=discriminator,
                      metadata=self.metadata)
        ctgan.compile(generator_optimizer=keras.optimizers.Adam(),
                      discriminator_optimizer=keras.optimizers.Adam())
        ctgan.build((16, self.data_dim))
        if keras.backend.backend() == "jax":
            ctgan.build_optimizers()
        logs = ctgan.fit(self.dataset, epochs=3)



import keras
from keras import ops

from teras._src.layers.layer_list import LayerList
from teras._src.layers.tabnet.attentive_transformer import \
    TabNetAttentiveTransformer
from teras._src.layers.tabnet.feature_transformer import \
    TabNetFeatureTransformer
from teras._src.api_export import teras_export


@teras_export("teras.models.TabNetEncoderBackbone")
class TabNetEncoderBackbone(keras.Model):
    """
    TabNetEncoder proposed by Arik et al. in the
    "TabNet: Attentive Interpretable Tabular Learning" paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        input_dim: int, The dimensionality of the input dataset.
        feature_transformer_dim: int, the dimensionality of the hidden
            representation in feature transformation block.
            The dense layer inside the `TabNetFeatureTransformer` first
            maps  the inputs to `feature_transformer_dim` * 2 dimension
            hidden representations and later the glu activation
            maps the hidden representations to
            `feature_transformer_dim` dimensions.
        decision_step_dim: int, the dimensionality of output at each
            decision step, which is later mapped to the final
            classification or regression output.
            It is recommended to keep `decision_step_dim` and
            `feature_transformer_dim` equal to each other.
            Adjusting these two parameters values is a good way of
            obtaining a tradeoff between performance and complexity.
        num_decision_steps: int, the number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset,
            the value tends to be higher. That said, a very high value
            of `num_decision_steps` may suffer from overfitting.
            Defaults to 5.
        num_shared_layers: int, Number of shared layers to use in the
            ``TabNetFeatureTransformer``.
            These shared layers are shared across decision steps.
            Defaults to 2.
        num_decision_dependent_layers: int, number of decision dependent
            layers to use in the `TabNetFeatureTransformer`.
            In simple words, `num_decision_dependent_layers` are created
            for each decision step in the `num_decision_steps`.
            For instance, if `num_decision_steps = 5` and
            `num_decision_dependent_layers = 2`
            then 10 layers will be created, 2 for each decision step.
        relaxation_factor: float, relaxation factor that promotes the
            reuse of each feature at different decision steps. When it
            is 1, a feature is enforced to be used only at one decision
            step and as it increases, more flexibility is provided to
            use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on
            the performance. Typically, a larger value for
            `num_decision_steps` favors for a larger `relaxation_factor`.
        batch_momentum: float, momentum value to use for
            `BatchNormalization` layer.
            Defaults to 0.9
        epsilon: float, epsilon is a small number for numerical stability
            during the computation of entropy loss.
            Defaults to 0.00001
        reuse_shared_layers: bool, whether to reset shared layers of the
            `TabNetFeatureTransformer` layer.
            Although we want to use the same shared layers across
            multiple instances of `TabNetFeatureTransformer` but we may
            not want to use the same shared layers across different
            `TabNetEncoder` instances.
            Defaults to `True`.
    """
    def __init__(self,
                 input_dim: int,
                 feature_transformer_dim: int,
                 decision_step_dim: int,
                 num_decision_steps: int = 5,
                 num_shared_layers: int = 2,
                 num_decision_dependent_layers: int = 5,
                 relaxation_factor: float = 1.5,
                 batch_momentum: float = 0.9,
                 epsilon: float = 1e-5,
                 reset_shared_layers: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.feature_transformer_dim = feature_transformer_dim
        self.decision_step_dim = decision_step_dim
        self.num_decision_steps = num_decision_steps
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.relaxation_factor = relaxation_factor
        self.batch_momentum = batch_momentum
        self.epsilon = epsilon
        self.reset_shared_layers = reset_shared_layers

        if self.reset_shared_layers:
            TabNetFeatureTransformer.reset_shared_layers()

        self.feature_transformers = LayerList([
            TabNetFeatureTransformer(
                hidden_dim=self.feature_transformer_dim,
                num_shared_layers=self.num_decision_steps,
                num_decision_dependent_layers=self.num_decision_dependent_layers,
                batch_momentum=self.batch_momentum,
                name=f"encoder_feature_transformer_{i}"
            )
            for i in range(self.num_decision_steps)
        ],
            sequential=False,
            name="encoder_feature_transformers"
        )
        self.attentive_transformers = LayerList([
            TabNetAttentiveTransformer(
                data_dim=input_dim,
                batch_momentum=self.batch_momentum,
                name=f"encoder_attentive_transformer_{i}"
            )
            for i in range(self.num_decision_steps - 1)
        ],
            sequential=False,
            name="encoder_attentive_transformers"
        )
        self.batch_norm = keras.layers.BatchNormalization(
            momentum=self.batch_momentum
        )

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self._input_shape = input_shape
        self.batch_norm.build(input_shape)
        self.feature_transformers.build(input_shape)
        input_shape = input_shape[:-1] + (
            self.feature_transformer_dim - self.decision_step_dim,)
        self.attentive_transformers.build(input_shape)

    def call(self, inputs, mask=None):
        normalized_inputs = self.batch_norm(inputs)
        batch_size = ops.shape(inputs)[0]
        decision_out_aggregated = ops.zeros(
            (batch_size, self.decision_step_dim))
        # During pretraining, we pass mask alongside inputs to the encoder
        masked_features = normalized_inputs
        if mask is not None:
            mask_values = mask
        else:
            mask_values = ops.zeros_like(inputs)
        aggregated_mask_values = ops.zeros_like(inputs)
        # Prior scales denote how much a feature has been used previously
        prior_scales = ops.ones_like(inputs)
        total_entropy = 0.

        for d_step in range(self.num_decision_steps):
            feat_transformer_out = self.feature_transformers[d_step](
                masked_features)
            if d_step > 0 or self.num_decision_steps == 1:
                decision_out = ops.relu(
                    feat_transformer_out[:, :self.decision_step_dim])
                decision_out_aggregated += decision_out

                # Aggregated masks are used for the visualization of the
                # feature importance attributes.
                scale = ops.sum(
                    decision_out, axis=-1, keepdims=True) / (
                                self.num_decision_steps - 1)
                aggregated_mask_values += mask_values / scale

            the_other_output_half = feat_transformer_out[:, self.decision_step_dim:]

            if d_step < self.num_decision_steps - 1:
                mask = self.attentive_transformers[d_step](
                    the_other_output_half,
                    prior_scales)
                # Relaxation factor controls the amount of reuse of
                # features between different decision blocks and updated
                # with the values of coefficients.
                prior_scales = prior_scales * (self.relaxation_factor -
                                               mask)

                # Entropy is used to penalize the amount of sparsity in
                # feature selection.
                total_entropy = total_entropy + ops.mean(
                    ops.sum(-mask * ops.log(mask + self.epsilon), axis=-1)
                ) / self.num_decision_steps - 1

                masked_features = mask * normalized_inputs

        self.add_loss(total_entropy)

        return decision_out_aggregated

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.decision_step_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            'input_dim': self.input_dim,
            'feature_transformer_dim': self.feature_transformer_dim,
            'decision_step_dim': self.decision_step_dim,
            'num_decision_steps': self.num_decision_steps,
            'num_shared_layers': self.num_shared_layers,
            'num_decision_dependent_layers': self.num_decision_dependent_layers,
            'relaxation_factor': self.relaxation_factor,
            'batch_momentum': self.batch_momentum,
            'epsilon': self.epsilon,
            'reset_shared_layers': self.reset_shared_layers,
        })
        return config

    def get_build_config(self):
        build_config = dict(input_shape=self._input_shape)
        return build_config

    def build_from_config(self, build_config):
        self.build(build_config["input_shape"])



import pytest
from keras.backend import backend


skip_on_torch = pytest.mark.skipif(
    backend() == "torch", reason="Test enters infinite loop on `torch` backend."
)

skip_for_now = pytest.mark.skip(reason="skip for now. will figure it out.")



import keras
import numpy as np

from teras._src.layers.categorical_extraction import CategoricalExtraction
from teras._src.layers.continuous_extraction import ContinuousExtraction
from teras._src.layers.tab_transformer.column_embedding import \
    TabTransformerColumnEmbedding
from teras._src.models.backbones.backbone import Backbone
from teras._src.models.backbones.transformer.encoder import \
    TransformerEncoderBackbone
from teras._src.api_export import teras_export


@teras_export("teras.models.TabTransformerBackbone")
class TabTransformerBackbone(Backbone):
    """
    TabTransformer backbone based on the architecture proposed in the
    "TabTransformer: Tabular Data Modeling Using Contextual Embeddings".

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        input_dim: int, dimensionality of the input dataset. i.e. the
            number of features in the dataset.
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use any value <=0 as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, dimensionality of the embeddings used
            by the model. It is also referred to as the `d_model` or
            model dimensionality.
        use_shared_embedding: bool, whether to use the shared embeddings.
            Defaults to `True`.
            If `False`, this layer will be effectively equivalent to a
            `keras.layers.Embedding` layer.
        shared_embedding_dim: int, dimensionality of the shared embeddings.
            Shared embeddings are the embeddings of the unique column
            identifiers, which according to the paper, help the model
            distinguish categories of one feature from the other.
            By default, its value is set to `embedding_dim / 8` as this
            setup is the most superior in the results presented by the authors.
        join_method: str, one of ['concat', 'add'] method to join the
            shared embeddings with feature embeddings.
            Defaults to 'concat', which is the recommended method,
            in which shared embeddings of `shared_embedding_dim` are
            concatenated with `embedding_dim - shared_embedding_dim`
            dimension feature embeddings.
            In 'add', shared embeddings have the same dimensions as the
            features, i.e. the `embedding_dim` and they are element-wise
            added to the features.
        num_layers: int, number of `TransformerEncoderLayer`s to use in
            the encoder.
        num_heads: int, number of attention heads to use in the
            `MultiHeadAttention` layer.
        feedforward_dim: int, hidden dimensionality to use in the
            `TransformerFeedForward` layer.
        attention_dropout: float, dropout value to use in the
            `MultiHeadAttention` layer. Defaults to 0.
        feedforward_dropout: float, dropout value to use in the
            `TransformerFeedForward` layer. Defaults to 0.
        layer_norm_epsilon: float, epsilon value to use in the
            `LayerNormalization` layer. Defaults to 1e-5.

    Shapes:
        input_shape: (batch_size, input_dim)

        output_shape: (batch_size, input_dim)
    """
    def __init__(self,
                 input_dim: int,
                 cardinalities: list,
                 embedding_dim: int,
                 use_shared_embedding: bool = True,
                 shared_embedding_dim: int = None,
                 join_method: str = "concat",
                 num_layers: int = 6,
                 num_heads: int = 8,
                 feedforward_dim: int = None,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 layer_norm_epsilon: float = 1e-5,
                 **kwargs):
        inputs = keras.layers.Input((input_dim,), name="inputs")
        categorical_idx = np.flatnonzero(np.array(cardinalities) != 0)
        continuous_idx = np.flatnonzero(np.array(cardinalities) == 0)

        # Deal with categorical data
        x_cat = CategoricalExtraction(categorical_idx)(inputs)
        x_cat = TabTransformerColumnEmbedding(
            cardinalities=cardinalities,
            embedding_dim=embedding_dim,
            use_shared_embedding=use_shared_embedding,
            shared_embedding_dim=shared_embedding_dim,
            join_method=join_method)(x_cat)
        x_cat = TransformerEncoderBackbone(
            input_dim=len(categorical_idx),
            embedding_dim=embedding_dim,
            num_layers=num_layers,
            num_heads=num_heads,
            feedforward_dim=feedforward_dim,
            attention_dropout=attention_dropout,
            feedforward_dropout=feedforward_dropout,
            layer_norm_epsilon=layer_norm_epsilon)(x_cat)
        x_cat = keras.layers.Flatten()(x_cat)

        # Deal with continuous data
        x_cont = ContinuousExtraction(continuous_idx)(inputs)
        x_cont = keras.layers.LayerNormalization(
            epsilon=layer_norm_epsilon)(x_cont)

        # Concatenate
        outputs = keras.layers.Concatenate()([x_cat, x_cont])
        super().__init__(inputs=inputs, outputs=outputs, **kwargs)

        self.input_dim = input_dim
        self.cardinalities = cardinalities
        self.embedding_dim = embedding_dim
        self.use_shared_embedding = use_shared_embedding
        self.shared_embedding_dim = shared_embedding_dim
        self.join_method = join_method
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.feedforward_dim = feedforward_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.layer_norm_epsilon = layer_norm_epsilon

    def compute_output_shape(self, input_shape):
        cards = np.array(self.cardinalities)
        continuous_dim = sum(cards == 0)
        categorical_dim = sum(cards != 0) * self.embedding_dim
        return input_shape[:1] + (continuous_dim + categorical_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "input_dim": self.input_dim,
            "cardinalities": self.cardinalities,
            "embedding_dim": self.embedding_dim,
            "use_shared_embedding": self.use_shared_embedding,
            "shared_embedding_dim": self.shared_embedding_dim,
            "join_method": self.join_method,
            "num_layers": self.num_layers,
            "num_heads": self.num_heads,
            "feedforward_dim": self.feedforward_dim,
            "attention_dropout": self.attention_dropout,
            "feedforward_dropout": self.feedforward_dropout,
            "layer_norm_epsilon": self.layer_norm_epsilon
        })
        return config



import os

import keras
import pytest
from keras import backend
from keras import random, ops
from keras.src.testing.test_case import TestCase
from teras._src.preprocessing.data_samplers.ctgan import CTGANDataSampler
from teras._src.preprocessing.data_transformers.ctgan import CTGANDataTransformer
from teras._src.testing.markers import skip_on_torch

from teras._src.models.gans.ctgan.generator import CTGANGenerator
from teras._src.utils import generate_fake_gemstone_data

if backend.backend() == "torch":
    pytest.skip("Tests enter inifite loop on torch backend.",
                allow_module_level=True)


class CTGANGeneratorTest(TestCase):
    def setUp(self):
        fake_gem_df = generate_fake_gemstone_data(num_samples=32)
        num_cols = ["depth", "table"]
        cat_cols = ["cut", "color", "clarity"]
        data_transformer = CTGANDataTransformer(continuous_features=num_cols,
                                                categorical_features=cat_cols)
        x_transformed = data_transformer.fit_transform(fake_gem_df)

        data_sampler = CTGANDataSampler(batch_size=8,
                                        categorical_features=cat_cols,
                                        metadata=data_transformer.get_metadata())
        self.dataset = data_sampler.get_dataset(x_transformed=x_transformed,
                                                x_original=fake_gem_df)
        self.metadata = data_transformer.get_metadata()
        self.data_dim = data_sampler.data_dim

    def test_valid_call(self):
        z = random.uniform((8, 16))
        generator = CTGANGenerator(data_dim=self.data_dim,
                                   metadata=self.metadata)
        outputs = generator(z)

    def test_valid_output_shape(self):
        z = random.uniform((8, 16))
        generator = CTGANGenerator(data_dim=self.data_dim,
                                   metadata=self.metadata)
        outputs = generator(z)
        self.assertEqual(ops.shape(outputs), (8, self.data_dim))

    @skip_on_torch
    def test_save_and_load(self):
        z = random.uniform((8, 16))
        generator = CTGANGenerator(data_dim=self.data_dim,
                                   metadata=self.metadata)
        outputs_original = generator(z)
        save_path = os.path.join(self.get_temp_dir(),
                                 "ctgan_generator.keras")
        generator.save(save_path)
        reloaded_generator = keras.models.load_model(save_path)
        self.assertIsInstance(reloaded_generator, CTGANGenerator)
        outputs_reloaded = reloaded_generator(z)
        self.assertAllClose(ops.convert_to_numpy(outputs_original),
                            ops.convert_to_numpy(outputs_reloaded))



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.losses.ctgan_generator_loss")
def ctgan_generator_loss(x_generated,
                         y_pred_generated,
                         cond_vectors,
                         mask,
                         metadata):
    """
    Loss for the Generator model in the CTGAN architecture.

    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        x_generated: Samples drawn from the input dataset
        y_pred_generated: Discriminator's output for the generated samples
        cond_vectors: Conditional vectors that are used for and with
            generated samples
        mask: Mask created during the conditional vectors generation step
        metadata: dict, metadata computed during the data transformation step.

    Returns:
        Generator's loss.
    """
    loss = []
    cross_entropy_loss = keras.losses.SparseCategoricalCrossentropy(
        from_logits=True, reduction=None)
    numerical_features_relative_indices = metadata["numerical"]["relative_indices_all"]
    features_relative_indices_all = metadata["relative_indices_all"]
    num_categories_all = metadata["categorical"]["num_categories_all"]
    # the first k features in the data are numerical which we'll ignore as
    # we're only concerned with the categorical features here
    offset = len(numerical_features_relative_indices)
    for i, index in enumerate(features_relative_indices_all[offset:]):
        logits = x_generated[:, index: index + num_categories_all[i]]
        temp_cond_vector = cond_vectors[:, i: i + num_categories_all[i]]
        labels = ops.argmax(temp_cond_vector, axis=1)
        ce_loss = cross_entropy_loss(y_pred=logits,
                                     y_true=labels
                                     )
        loss.append(ce_loss)
    loss = ops.stack(loss, axis=1)
    loss = ops.sum(loss * ops.cast(mask, dtype="float32")
                   ) / ops.cast(ops.shape(y_pred_generated)[0], dtype="float32")
    loss = -ops.mean(y_pred_generated) * loss
    return loss


@teras_export("teras.losses.ctgan_discriminator_loss")
def ctgan_discriminator_loss(y_pred_real, y_pred_generated):
    """
    Loss for the Discriminator model in the CTGAN architecture.

    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        y_pred_real: Discriminator's output for real samples
        y_pred_generated: Discriminator's output for generated samples

    Returns:
        Discriminator's loss.
    """
    return -(ops.mean(y_pred_real) - ops.mean(y_pred_generated))
